# in_memory_vfs.py
#
# LICENSE
#
# The MIT License
#
# Copyright (c) 2024 TileDB, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# DESCRIPTION
#
# When run, this program will create a random numpy array, create a TileDB
# DenseArray from it in memory, and read the entire array back from memory.
# It then compares the original numpy array with the TileDB array to check
# that data was written and read correctly (prints True) or not (prints False).
#

import numpy as np

import tiledb

# Create TileDB VFS
vfs = tiledb.VFS()

file = "mem://myarray"
data = np.random.rand(10, 10)

if not vfs.is_file(file):
    with tiledb.from_numpy(file, data):
        pass

with tiledb.open(file) as A:
    print(np.all(A[:] == data))



#
# This file generates the sidebar/toctree for all TileDB projects and should
# be copied to each project when it is updated.
#
# This file is originally from the RobotPy documentation project
# https://github.com/robotpy/robotpy-docs, licensed under Apache v2.
#

import os


def write_if_changed(fname, contents):

    try:
        with open(fname, "r") as fp:
            old_contents = fp.read()
    except:
        old_contents = ""

    if old_contents != contents:
        with open(fname, "w") as fp:
            fp.write(contents)


def generate_sidebar(conf, conf_api):

    version = conf["rtd_version"]

    lines = [
        "",
        ".. DO NOT MODIFY! THIS PAGE IS AUTOGENERATED!",
        "   To edit the sidebar, modify gensidebar.py and re-build the docs.",
        "",
    ]

    url_base = "https://tiledb-inc-tiledb.readthedocs-hosted.com"
    lang = "en"

    def toctree(name):
        lines.extend(
            [".. toctree::", "    :caption: %s" % name, "    :maxdepth: 1", ""]
        )

    def endl():
        lines.append("")

    def write(desc, link):
        if conf_api == "tiledb":
            args = desc, link
        else:
            args = desc, "%s/%s/%s/%s.html" % (url_base, lang, version, link)

        lines.append("    %s <%s>" % args)

    def write_api(project, desc, rst_page):
        # From non-root project to root project link
        if project == "tiledb" and conf_api != "tiledb":
            args = desc, url_base, lang, version, rst_page
            lines.append("    %s API <%s/%s/%s/%s.html>" % args)
        # From anything to non-root project link
        elif project != conf_api:
            args = desc, url_base, project, lang, version, rst_page
            lines.append("    %s API <%s/projects/%s/%s/%s/%s.html>" % args)
        # Local project link
        else:
            args = desc, rst_page
            lines.append("    %s API <%s>" % args)

    def write_api_url(desc, url):
        lines.append("    %s API <%s>" % (desc, url))

    #
    # Specify the sidebar contents here
    #

    toctree("API Reference")
    write_api("tiledb", "C", "c-api")
    write_api("tiledb", "C++", "c++-api")
    write_api("tiledb-py", "Python", "python-api")
    write_api_url("R", "https://tiledb-inc.github.io/TileDB-R/reference/index.html")
    write_api_url("Java", "https://www.javadoc.io/doc/io.tiledb/tiledb-java")
    write_api_url("Go", "https://godoc.org/github.com/TileDB-Inc/TileDB-Go")
    endl()

    write_if_changed("_sidebar.rst.inc", "\n".join(lines))



import shutil
import tempfile

import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_array_equal

import tiledb
import tiledb.main as main


@pytest.mark.skipif(
    not main.test_webp_filter.webp_filter_exists(),
    reason="Can't create WebP filter; built with TILEDB_WEBP=OFF",
)
@pytest.mark.parametrize(
    "format, quality, lossless",
    [
        (
            tiledb.filter.WebpFilter.WebpInputFormat.WEBP_RGB,
            100.0,
            False,
        ),  # Test setting format with enum values
        (tiledb.filter.WebpFilter.WebpInputFormat.WEBP_BGR, 50.0, True),
        (tiledb.filter.WebpFilter.WebpInputFormat.WEBP_RGBA, 25.5, False),
        (4, 0.0, True),  # Test setting format with integral type
    ],
)
def test_webp_ctor(format, quality, lossless):
    webp_filter = tiledb.WebpFilter(
        input_format=format, quality=quality, lossless=lossless
    )
    np.testing.assert_equal(
        webp_filter.input_format, tiledb.filter.WebpFilter.WebpInputFormat(format)
    )
    np.testing.assert_equal(webp_filter.quality, quality)
    np.testing.assert_equal(webp_filter.lossless, lossless)


@pytest.mark.skipif(
    not main.test_webp_filter.webp_filter_exists(),
    reason="Can't create WebP filter; built with TILEDB_WEBP=OFF",
)
@pytest.mark.parametrize(
    "attr_dtype, dim_dtype, var, sparse",
    [
        (np.int64, np.int64, None, True),  # Sparse arrays are not supported
        (np.int64, np.int64, True, False),  # Variable attributes are not supported
    ],
)
def test_webp_init(attr_dtype, dim_dtype, var, sparse):
    with pytest.raises(tiledb.TileDBError):
        tiledb.ArraySchema(
            domain=tiledb.Domain(
                [
                    tiledb.Dim("y", domain=(1, 100), dtype=dim_dtype),
                    tiledb.Dim("x", domain=(1, 300), dtype=dim_dtype),
                ]
            ),
            attrs=[
                tiledb.Attr(
                    "rgb", dtype=attr_dtype, var=var, filters=[tiledb.WebpFilter()]
                )
            ],
            sparse=sparse,
        )


def make_image_data(width, height, pixel_depth):
    center_x = width / 2
    center_y = height / 2

    colors = {
        "red": [255, 0, 0],
        "green": [0, 255, 0],
        "blue": [0, 0, 255],
        "white": [255, 255, 255],
        "black": [0, 0, 0],
    }
    if pixel_depth > 3:
        for color in colors.values():
            color.append(255)

    rgb = []
    for row in range(0, height):
        r = []
        for col in range(0, width):
            if row < center_y and col < center_x:
                r.append(colors["red"])
            elif row < center_y and col > center_x:
                r.append(colors["green"])
            elif row > center_y and col < center_x:
                r.append(colors["blue"])
            elif row > center_y and col > center_x:
                r.append(colors["white"])
            elif row == center_y or col == center_x:
                r.append(colors["black"])
        rgb.append(r)
    return rgb


@pytest.mark.skipif(
    not main.test_webp_filter.webp_filter_exists(),
    reason="Can't create WebP filter; built with TILEDB_WEBP=OFF",
)
@pytest.mark.parametrize(
    "width, height",
    [
        (3, 7),
        (20, 20),
        (40, 40),
        (479, 149),
        (1213, 1357),
        (1111, 3333),
    ],
)
@pytest.mark.parametrize(
    "colorspace",
    [
        tiledb.filter.WebpFilter.WebpInputFormat.WEBP_RGB,
        tiledb.filter.WebpFilter.WebpInputFormat.WEBP_BGR,
        tiledb.filter.WebpFilter.WebpInputFormat.WEBP_RGBA,
        tiledb.filter.WebpFilter.WebpInputFormat.WEBP_BGRA,
    ],
)
@pytest.mark.parametrize("lossless", [True, False])
def test_webp_filter(width, height, colorspace, lossless):
    pixel_depth = (
        3
        if int(colorspace) < int(tiledb.filter.WebpFilter.WebpInputFormat.WEBP_RGBA)
        else 4
    )
    data = make_image_data(width, height, pixel_depth)
    data = np.array(data, dtype=np.uint8).reshape(height, width * pixel_depth)

    y_tile = round(height / 2)
    x_tile = round(width / 2) * pixel_depth

    dim_dtype = np.min_scalar_type(data.size)
    dims = (
        tiledb.Dim(
            name="Y",
            domain=(1, height),
            dtype=dim_dtype,
            tile=y_tile,
        ),
        tiledb.Dim(
            name="X",
            domain=(1, width * pixel_depth),
            dtype=dim_dtype,
            tile=x_tile,
        ),
    )
    schema = tiledb.ArraySchema(
        domain=tiledb.Domain(*dims),
        attrs=[
            tiledb.Attr(
                name="rgb",
                dtype=np.uint8,
                filters=[
                    tiledb.WebpFilter(
                        input_format=colorspace, quality=100.0, lossless=lossless
                    )
                ],
            )
        ],
    )

    uri = tempfile.mkdtemp()
    tiledb.Array.create(uri, schema)
    with tiledb.open(uri, "w") as A:
        A[:] = data
    with tiledb.open(uri, "r") as A:
        read_image = A[:]

    if lossless:
        assert_array_equal(
            data.reshape(np.array(read_image["rgb"]).shape), read_image["rgb"]
        )
    else:
        assert_allclose(
            data.reshape(np.array(read_image["rgb"]).shape), read_image["rgb"], 125
        )

    # Cleanup.
    shutil.rmtree(uri)



# query_condition_dense.py
#
# LICENSE
#
# The MIT License
#
# Copyright (c) 2021 TileDB, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

# This example creates an array with one string-typed attribute,
# writes sample data to the array, and then prints out a filtered
# dataframe using the TileDB QueryCondition feature.

from pprint import pprint

import numpy as np

import tiledb

uri = "query_condition_dense"


def create_array(path):
    # create a dense array
    dom = tiledb.Domain(
        tiledb.Dim(name="coords", domain=(1, 10), tile=1, dtype=np.uint32)
    )
    attrs = [
        tiledb.Attr(name="attr1", dtype=np.uint64),
        tiledb.Attr(name="attr2", dtype=np.float64),
    ]
    schema = tiledb.ArraySchema(domain=dom, attrs=attrs, sparse=False)
    tiledb.Array.create(path, schema, overwrite=True)

    # fill array with randomized values
    with tiledb.open(path, "w") as arr:
        rand = np.random.default_rng()
        arr[:] = {
            "attr1": rand.integers(low=0, high=10, size=10),
            "attr2": rand.random(size=10),
        }


def read_array(path):
    with tiledb.open(uri) as arr:
        print("--- without query condition:")
        print()
        pprint(arr[:])
        print()

    with tiledb.open(uri) as arr:
        qc = "(2 < attr1 < 6) and (attr2 < 0.5 or attr2 > 0.85)"
        print(f"--- with query condition {qc}:")

        print(f"--- the fill value for attr1 is {arr.attr('attr1').fill}")
        print(f"--- the fill value for attr2 is {arr.attr('attr2').fill}")

        print()
        res = arr.query(cond=qc)[:]
        pprint(res)


if __name__ == "__main__":
    """Example output for `python query_condition_dense.py`:

    --- without query condition:

    OrderedDict([('attr1', array([4, 0, 9, 7, 6, 0, 0, 5, 7, 5], dtype=uint64)),
                ('attr2',
                array([0.74476144, 0.47211544, 0.99054245, 0.36640416, 0.91699594,
        0.06216043, 0.58581863, 0.00505695, 0.7486192 , 0.87649422]))])

    --- with query condition (2 < attr1 < 6) and (attr2 < 0.5 or attr2 > 0.85):
    --- the fill value for attr1 is [18446744073709551615]
    --- the fill value for attr2 is [nan]

    OrderedDict([('attr1',
                array([18446744073709551615, 18446744073709551615, 18446744073709551615,
        18446744073709551615, 18446744073709551615, 18446744073709551615,
        18446744073709551615,                    5, 18446744073709551615,
                            5], dtype=uint64)),
                ('attr2',
                array([       nan,        nan,        nan,        nan,        nan,
                nan,        nan, 0.00505695,        nan, 0.87649422]))])
    """
    create_array(uri)
    read_array(uri)



import tempfile
from pathlib import Path

import numpy as np
from numpy.testing import assert_array_equal

import tiledb

from .common import DiskTestCase


class UtilTest(DiskTestCase):
    def test_empty_like(self):
        arr = np.zeros((10, 10), dtype=np.float32)

        def check_schema(self, s):
            self.assertEqual(s.attr(0).dtype, np.float32)
            self.assertEqual(s.shape, (10, 10))
            self.assertEqual(s.domain.dim(0).shape, (10,))
            self.assertEqual(s.domain.dim(1).shape, (10,))

        with self.assertRaises(ValueError):
            tiledb.schema_like("", None)

        schema = tiledb.schema_like(arr, tile=1)
        self.assertIsInstance(schema, tiledb.ArraySchema)
        check_schema(self, schema)

        uri = self.path("empty_like")
        T = tiledb.empty_like(uri, arr)
        check_schema(self, T.schema)
        self.assertEqual(T.shape, arr.shape)
        self.assertEqual(T.dtype, arr.dtype)

        uri = self.path("empty_like_shape")
        T = tiledb.empty_like(uri, arr.shape, dtype=arr.dtype)
        check_schema(self, T.schema)
        self.assertEqual(T.shape, arr.shape)
        self.assertEqual(T.dtype, arr.dtype)

        # test a fake object with .shape, .ndim, .dtype
        class FakeArray(object):
            def __init__(self, shape, dtype):
                self.shape = shape
                self.ndim = len(shape)
                self.dtype = dtype

        fake = FakeArray((3, 3), np.int16)
        schema2 = tiledb.empty_like(self.path("fake_like"), fake)
        self.assertIsInstance(schema2, tiledb.Array)
        self.assertEqual(schema2.shape, fake.shape)
        self.assertEqual(schema2.dtype, fake.dtype)
        self.assertEqual(schema2.ndim, fake.ndim)

        # test passing shape and dtype directly
        schema3 = tiledb.schema_like(shape=(4, 4), dtype=np.float32)
        self.assertIsInstance(schema3, tiledb.ArraySchema)
        self.assertEqual(schema3.attr(0).dtype, np.float32)
        self.assertEqual(schema3.domain.dim(0).tile, 4)
        schema3 = tiledb.schema_like(shape=(4, 4), dtype=np.float32, tile=1)
        self.assertEqual(schema3.domain.dim(0).tile, 1)

    def test_open(self):
        uri = self.path("load")
        with tiledb.from_numpy(uri, np.array(np.arange(3))) as T:
            with tiledb.open(uri) as T2:
                self.assertEqual(T.schema, T2.schema)
                assert_array_equal(T, T2)

    def test_save(self):
        uri = self.path("test_save")
        arr = np.array(np.arange(3))
        with tiledb.save(uri, arr):
            with tiledb.open(uri) as T:
                assert_array_equal(arr, T)

    def test_array_exists(self):
        with tempfile.NamedTemporaryFile() as tmpfn:
            self.assertFalse(tiledb.array_exists(tmpfn.name))

        uri = self.path("test_array_exists_dense")
        with tiledb.from_numpy(uri, np.arange(0, 5)) as T:
            self.assertTrue(tiledb.array_exists(uri))
            self.assertTrue(tiledb.array_exists(uri, isdense=True))
            self.assertFalse(tiledb.array_exists(uri, issparse=True))

        uri = self.path("test_array_exists_sparse")
        dom = tiledb.Domain(tiledb.Dim(domain=(0, 3), tile=4, dtype=int))
        att = tiledb.Attr(dtype=int)
        schema = tiledb.ArraySchema(domain=dom, attrs=(att,), sparse=True)
        tiledb.Array.create(uri, schema)

        with tiledb.SparseArray(uri, mode="w") as T:
            T[[0, 1]] = np.array([0, 1])

        self.assertTrue(tiledb.array_exists(uri))
        self.assertTrue(tiledb.array_exists(uri, issparse=True))
        self.assertFalse(tiledb.array_exists(uri, isdense=True))

        uri3 = self.path("test_array_exists_deleted")
        with tiledb.from_numpy(uri3, np.arange(0, 5)) as T:
            self.assertTrue(tiledb.array_exists(uri3))
        tiledb.Array.delete_array(uri3)
        self.assertFalse(tiledb.array_exists(uri3))

        # test with context
        ctx = tiledb.Ctx()
        self.assertFalse(tiledb.array_exists(uri3, ctx=ctx))
        with tiledb.from_numpy(uri3, np.arange(0, 5), ctx=ctx) as T:
            self.assertTrue(tiledb.array_exists(uri3, ctx=ctx))

    def test_ls(self):
        def create_array(array_name, sparse):
            dom = tiledb.Domain(
                tiledb.Dim(name="rows", domain=(1, 4), tile=4, dtype=np.int32),
                tiledb.Dim(name="cols", domain=(1, 4), tile=4, dtype=np.int32),
            )
            schema = tiledb.ArraySchema(
                domain=dom, sparse=sparse, attrs=[tiledb.Attr(name="a", dtype=np.int32)]
            )
            tiledb.Array.create(array_name, schema)

        uri = tempfile.mkdtemp()

        tiledb.group_create(str(Path(uri) / "my_group"))
        tiledb.group_create(str(Path(uri) / "my_group" / "dense_arrays"))
        tiledb.group_create(str(Path(uri) / "my_group" / "sparse_arrays"))

        create_array(str(Path(uri) / "my_group" / "dense_arrays" / "array_A"), False)
        create_array(str(Path(uri) / "my_group" / "dense_arrays" / "array_B"), False)
        create_array(str(Path(uri) / "my_group" / "sparse_arrays" / "array_C"), True)
        create_array(str(Path(uri) / "my_group" / "sparse_arrays" / "array_D"), True)

        group_uri = "{}/my_group".format(uri)
        # List children
        results = []
        tiledb.ls(
            group_uri,
            lambda obj_path, obj_type: results.append((Path(obj_path).name, obj_type)),
        )
        self.assertEqual(
            results,
            [
                ("dense_arrays", "group"),
                ("sparse_arrays", "group"),
            ],
        )

        # List children of a group
        dense_arrays_uri = "{}/my_group/dense_arrays".format(uri)
        results = []
        tiledb.ls(
            dense_arrays_uri,
            lambda obj_path, obj_type: results.append((Path(obj_path).name, obj_type)),
        )
        self.assertEqual(
            results,
            [
                ("array_A", "array"),
                ("array_B", "array"),
            ],
        )

        # test with a callback that always throws an exception to see if it is propagated
        with self.assertRaises(tiledb.TileDBError) as excinfo:
            tiledb.ls(dense_arrays_uri, lambda x, y: 1 / 0)
        assert "ZeroDivisionError: division by zero" in str(excinfo.value)

    def test_object_type(self):
        uri = self.path("test_object_type")

        # None case
        self.assertIsNone(tiledb.object_type(uri))

        # Array case
        with tiledb.from_numpy(uri, np.arange(0, 5)) as T:
            self.assertEqual(tiledb.object_type(uri), "array")
        tiledb.Array.delete_array(uri)

        # Group case
        tiledb.group_create(uri)
        self.assertEqual(tiledb.object_type(uri), "group")



from collections import OrderedDict

import numpy as np
import pytest

import tiledb

from .common import (
    DiskTestCase,
    assert_array_equal,
    assert_dict_arrays_equal,
)

SUPPORTED_INTEGER_DTYPES = (
    np.uint8,
    np.uint16,
    np.uint32,
    np.uint64,
    np.int8,
    np.int16,
    np.int32,
    np.int64,
)

SUPPORTED_DATETIME64_RESOLUTION = ("Y", "M", "W", "D", "h", "m", "s", "ms", "us", "ns")


@pytest.mark.parametrize("sparse", (True, False))
class TestReadSubarray1D(DiskTestCase):
    data1 = np.random.rand(101)
    data2 = np.random.randint(-1000, 1000, (101,), dtype=np.int16)
    label_data = np.linspace(-1.0, 1.0, 101)

    @pytest.fixture
    def array_uri(self, sparse):
        """Create TileDB array, write data, and return the URI."""
        suffix = "1d_label_sparse" if sparse else "1d_label_dense"
        uri = self.path(f"read_subarray_{suffix}")
        dim1 = tiledb.Dim(name="d1", domain=(0, 100), tile=101, dtype=np.int32)
        schema = tiledb.ArraySchema(
            domain=tiledb.Domain(dim1),
            attrs=[
                tiledb.Attr(name="a1", dtype=np.float64),
                tiledb.Attr(name="a2", dtype=np.int16),
            ],
            dim_labels={
                0: {
                    "l1": dim1.create_label_schema("increasing", np.float64),
                    "l2": dim1.create_label_schema("decreasing", np.float64),
                }
            },
            sparse=sparse,
        )
        tiledb.Array.create(uri, schema)
        data_buffers = {
            "a1": self.data1,
            "a2": self.data2,
            "l1": self.label_data,
            "l2": np.flip(self.label_data),
        }
        with tiledb.open(uri, "w") as array:
            if sparse:
                array[np.arange(101, dtype=np.int32)] = data_buffers
            else:
                array[...] = data_buffers
        return uri

    def test_read_full_array(self, array_uri):
        with tiledb.open(array_uri, "r") as array:
            sparse = array.schema.sparse
            subarray = tiledb.Subarray(array)
            subarray.add_dim_range(0, (0, 100))
            result = array.read_subarray(subarray)

        expected = OrderedDict()
        if sparse:
            expected["d1"] = np.arange(101, dtype=np.int32)
        expected["a1"] = self.data1
        expected["a2"] = self.data2

        assert_dict_arrays_equal(result, expected, not sparse)

    def test_read_partial(self, array_uri):
        with tiledb.open(array_uri, "r") as array:
            sparse = array.schema.sparse
            subarray = tiledb.Subarray(array)
            subarray.add_dim_range(0, (10, 20))
            result = array.read_subarray(subarray)

        expected = OrderedDict()
        if sparse:
            expected["d1"] = np.arange(10, 21, dtype=np.int32)
        expected["a1"] = self.data1[10:21]
        expected["a2"] = self.data2[10:21]

        assert_dict_arrays_equal(result, expected, not sparse)

    def test_read_multiple_ranges(self, array_uri):
        with tiledb.open(array_uri, "r") as array:
            sparse = array.schema.sparse
            subarray = tiledb.Subarray(array)
            subarray.add_dim_range(0, (3, 3))
            subarray.add_dim_range(0, (1, 2))
            subarray.add_dim_range(0, (5, 10))
            result = array.read_subarray(subarray)

        expected = OrderedDict()
        d1_expected = np.array([3, 1, 2, 5, 6, 7, 8, 9, 10], dtype=np.int32)
        if sparse:
            expected["d1"] = d1_expected
        expected["a1"] = self.data1[d1_expected]
        expected["a2"] = self.data2[d1_expected]

        assert_dict_arrays_equal(result, expected, not sparse)

    def test_read_single_attr(self, array_uri):
        with tiledb.open(array_uri, attr="a1", mode="r") as array:
            sparse = array.schema.sparse
            subarray = tiledb.Subarray(array)
            subarray.add_dim_range(0, (10, 20))
            result = array.read_subarray(subarray)

        expected = OrderedDict()
        if sparse:
            expected["d1"] = np.arange(10, 21, dtype=np.int32)
        expected["a1"] = self.data1[10:21]

        assert_dict_arrays_equal(result, expected, not sparse)

    def test_read_full_array_by_increasing_label(self, array_uri):
        with tiledb.open(array_uri, "r") as array:
            sparse = array.schema.sparse
            subarray = tiledb.Subarray(array)
            subarray.add_label_range("l1", (-1.0, 1.0))
            result = array.read_subarray(subarray)

        expected = OrderedDict()
        if sparse:
            expected["d1"] = np.arange(101, dtype=np.int32)
        expected["a1"] = self.data1
        expected["a2"] = self.data2

        assert_dict_arrays_equal(result, expected, not sparse)

    def test_read_partial_by_increasing_label(self, array_uri):
        with tiledb.open(array_uri, "r") as array:
            sparse = array.schema.sparse
            subarray = tiledb.Subarray(array)
            subarray.add_label_range("l1", (0.0, 1.0))
            result = array.read_subarray(subarray)

        expected = OrderedDict()
        if sparse:
            expected["d1"] = np.arange(50, 101, dtype=np.int32)
        expected["a1"] = self.data1[50:]
        expected["a2"] = self.data2[50:]

        assert_dict_arrays_equal(result, expected, not sparse)

    def test_read_partial_by_decreasing_label(self, array_uri):
        with tiledb.open(array_uri, "r") as array:
            sparse = array.schema.sparse
            subarray = tiledb.Subarray(array)
            subarray.add_label_range("l2", (0.0, 1.0))
            result = array.read_subarray(subarray)

        expected = OrderedDict()
        if sparse:
            expected["d1"] = np.arange(51, dtype=np.int32)
        expected["a1"] = self.data1[:51]
        expected["a2"] = self.data2[:51]

        assert_dict_arrays_equal(result, expected, not sparse)

    def test_read_by_label_no_data(self, array_uri):
        with tiledb.open(array_uri, "r") as array:
            subarray = tiledb.Subarray(array)
            subarray.add_label_range("l1", (0.01, 0.012))
            with pytest.raises(tiledb.TileDBError):
                array.read_subarray(subarray)


@pytest.mark.parametrize("dim_res", SUPPORTED_DATETIME64_RESOLUTION)
@pytest.mark.parametrize("sparse", (True, False))
class TestReadSubarrayDenseDatetime1D(DiskTestCase):
    data = np.random.rand(100)

    @pytest.fixture
    def array_uri(self, dim_res, sparse):
        """Create TileDB array, write data, and return the URI."""
        suffix = f"datetime_{dim_res}_sparse" if sparse else f"datetime_{dim_res}_dense"
        uri = self.path(f"read_subarray_1d_datetime_{suffix}")
        start_time = np.datetime64("2000-01-01", dim_res)
        domain = (start_time, start_time + np.timedelta64(99, dim_res))
        schema = tiledb.ArraySchema(
            tiledb.Domain(
                tiledb.Dim(
                    name="d1",
                    domain=domain,
                    tile=100,
                    dtype=np.dtype(f"M8[{dim_res}]"),
                )
            ),
            [tiledb.Attr(name="a1", dtype=np.float64)],
            sparse=sparse,
        )
        tiledb.Array.create(uri, schema)
        with tiledb.open(uri, "w") as array:
            if sparse:
                array[
                    np.arange(
                        domain[0],
                        domain[1] + np.timedelta64(1, dim_res),
                        np.timedelta64(1, dim_res),
                    )
                ] = self.data
            else:
                array[...] = self.data
        return uri

    def test_read_full_array(self, array_uri, dim_res):
        with tiledb.open(array_uri, "r") as array:
            sparse = array.schema.sparse
            subarray = tiledb.Subarray(array)
            start_time = np.datetime64("2000-01-01", dim_res)
            domain = (start_time, start_time + np.timedelta64(99, dim_res))
            subarray.add_dim_range(0, domain)
            result = array.read_subarray(subarray)

        expected = OrderedDict()
        if sparse:
            expected["d1"] = np.arange(
                start_time,
                start_time + np.timedelta64(100, dim_res),
                np.timedelta64(1, dim_res),
            )
        expected["a1"] = self.data

        assert_dict_arrays_equal(result, expected, not sparse)

    def test_read_partial(self, array_uri, dim_res):
        with tiledb.open(array_uri, "r") as array:
            sparse = array.schema.sparse
            subarray = tiledb.Subarray(array)
            start_time = np.datetime64("2000-01-01", dim_res)
            dim_range = (
                start_time + np.timedelta64(10, dim_res),
                start_time + np.timedelta64(20, dim_res),
            )
            subarray.add_dim_range(0, dim_range)
            result = array.read_subarray(subarray)
        assert_array_equal(result["a1"], self.data[10:21])

        expected = OrderedDict()
        if sparse:
            expected["d1"] = np.arange(
                start_time + np.timedelta64(10, dim_res),
                start_time + np.timedelta64(21, dim_res),
                np.timedelta64(1, dim_res),
            )
        expected["a1"] = self.data[10:21]

        assert_dict_arrays_equal(result, expected, not sparse)


@pytest.mark.parametrize("sparse", (True, False))
class TestReadSubarray2D(DiskTestCase):
    data_a1 = np.random.rand(16).reshape(4, 4)
    data_a2 = np.random.randint(-1000, 1000, (4, 4), dtype=np.int16)
    data_l1 = np.arange(-2, 2)
    data_l2 = np.arange(1, -3, -1)

    @pytest.fixture
    def array_uri(self, sparse):
        """Create TileDB array, write data, and return the URI."""
        suffix = "2d_sparse" if sparse else "2d_dense"
        uri = self.path(f"read_subarray_{suffix}")
        dim1 = tiledb.Dim(name="d1", domain=(0, 3), tile=4, dtype=np.int32)
        dim2 = tiledb.Dim(name="d2", domain=(0, 3), tile=4, dtype=np.int32)
        schema = tiledb.ArraySchema(
            domain=tiledb.Domain(dim1, dim2),
            attrs=[
                tiledb.Attr(name="a1", dtype=np.float64),
                tiledb.Attr(name="a2", dtype=np.int16),
            ],
            dim_labels={
                0: {"l1": dim1.create_label_schema("increasing", np.int32)},
                1: {"l2": dim1.create_label_schema("decreasing", np.int32)},
            },
            sparse=sparse,
        )
        tiledb.Array.create(uri, schema)
        if sparse:
            _schema = tiledb.ArraySchema.load(uri)
            with tiledb.open(_schema.dim_label("l1").uri, mode="w") as label1:
                label1[:] = self.data_l1
            with tiledb.open(_schema.dim_label("l2").uri, mode="w") as label2:
                label2[:] = self.data_l2
            data_d1, data_d2 = np.meshgrid(np.arange(4), np.arange(4), indexing="ij")
            with tiledb.open(uri, "w") as array:
                array[data_d1.flatten(), data_d2.flatten()] = {
                    "a1": self.data_a1,
                    "a2": self.data_a2,
                }
        else:
            with tiledb.open(uri, "w") as array:
                array[...] = {
                    "a1": self.data_a1,
                    "a2": self.data_a2,
                    "l1": self.data_l1,
                    "l2": self.data_l2,
                }

        return uri

    def test_read_full_array(self, array_uri):
        with tiledb.open(array_uri) as array:
            sparse = array.schema.sparse
            subarray = tiledb.Subarray(array)
            subarray.add_dim_range(0, (0, 3))
            subarray.add_dim_range(1, (0, 3))
            result = array.read_subarray(subarray)
        if sparse:
            # Construct the expected result
            data_d1, data_d2 = np.meshgrid(
                np.arange(4, dtype=np.int32),
                np.arange(4, dtype=np.int32),
                indexing="ij",
            )
            expected = {
                "d1": data_d1.flatten(),
                "d2": data_d2.flatten(),
                "a1": self.data_a1.flatten(),
                "a2": self.data_a2.flatten(),
            }
        else:
            expected = {"a1": self.data_a1, "a2": self.data_a2}
        assert_dict_arrays_equal(result, expected, not sparse)

    def test_read_mixed_ranges(self, array_uri):
        with tiledb.open(array_uri) as array:
            sparse = array.schema.sparse
            subarray = tiledb.Subarray(array)
            subarray.add_dim_range(0, (0, 1))
            result = array.read_subarray(subarray)
        if sparse:
            data_d1, data_d2 = np.meshgrid(
                np.arange(2, dtype=np.int32),
                np.arange(4, dtype=np.int32),
                indexing="ij",
            )
            expected = {
                "d1": data_d1.flatten(),
                "d2": data_d2.flatten(),
                "a1": self.data_a1[0:2, :].flatten(),
                "a2": self.data_a2[0:2, :].flatten(),
            }
        else:
            expected = {"a1": self.data_a1[0:2, :], "a2": self.data_a2[0:2, :]}
        assert_dict_arrays_equal(result, expected, not sparse)


@pytest.mark.parametrize("sparse", (True, False))
class TestReadSubarrayNegativeDomain2D(DiskTestCase):
    data_a1 = np.random.rand(121).reshape(11, 11)
    data_a2 = np.random.randint(-1000, 1000, (11, 11), dtype=np.int16)

    @pytest.fixture
    def array_uri(self, sparse):
        """Create TileDB array, write data, and return the URI."""
        suffix = "_sparse" if sparse else "_dense"
        uri = self.path(f"read_subarray_{suffix}")
        dim1 = tiledb.Dim(name="d1", domain=(-5, 5), tile=4, dtype=np.int32)
        dim2 = tiledb.Dim(name="d2", domain=(-5, 5), tile=4, dtype=np.int32)
        schema = tiledb.ArraySchema(
            domain=tiledb.Domain(dim1, dim2),
            attrs=[
                tiledb.Attr(name="a1", dtype=np.float64),
                tiledb.Attr(name="a2", dtype=np.int16),
            ],
            sparse=sparse,
        )
        tiledb.Array.create(uri, schema)
        if sparse:
            data_d1, data_d2 = np.meshgrid(
                np.arange(-5, 6), np.arange(-5, 6), indexing="ij"
            )
            with tiledb.open(uri, "w") as array:
                array[data_d1.flatten(), data_d2.flatten()] = {
                    "a1": self.data_a1,
                    "a2": self.data_a2,
                }
        else:
            with tiledb.open(uri, "w") as array:
                array[...] = {"a1": self.data_a1, "a2": self.data_a2}

        return uri

    def test_read_full_array(self, array_uri):
        with tiledb.open(array_uri) as array:
            sparse = array.schema.sparse
            subarray = tiledb.Subarray(array)
            subarray.add_dim_range(0, (-5, 5))
            subarray.add_dim_range(1, (-5, 5))
            result = array.read_subarray(subarray)
        if sparse:
            # Construct the expected result
            data_d1, data_d2 = np.meshgrid(
                np.arange(-5, 6, dtype=np.int32),
                np.arange(-5, 6, dtype=np.int32),
                indexing="ij",
            )
            expected = {
                "d1": data_d1.flatten(),
                "d2": data_d2.flatten(),
                "a1": self.data_a1.flatten(),
                "a2": self.data_a2.flatten(),
            }
        else:
            expected = {"a1": self.data_a1, "a2": self.data_a2}
        assert_dict_arrays_equal(result, expected, not sparse)

    def test_read_mixed_ranges(self, array_uri):
        with tiledb.open(array_uri) as array:
            sparse = array.schema.sparse
            subarray = tiledb.Subarray(array)
            subarray.add_dim_range(1, (-1, 2))
            result = array.read_subarray(subarray)

        if sparse:
            data_d1, data_d2 = np.meshgrid(
                np.arange(-5, 6, dtype=np.int32),
                np.arange(-1, 3, dtype=np.int32),
                indexing="ij",
            )
            expected = {
                "d1": data_d1.flatten(),
                "d2": data_d2.flatten(),
                "a1": self.data_a1[:, 4:8].flatten(),
                "a2": self.data_a2[:, 4:8].flatten(),
            }
        else:
            expected = {"a1": self.data_a1[:, 4:8], "a2": self.data_a2[:, 4:8]}
        assert_dict_arrays_equal(result, expected, not sparse)


class TestReadSubarraySparseArray1D(DiskTestCase):
    data_dim1 = np.linspace(-1.0, 1.0, 5)
    data_attr1 = np.arange(5, dtype=np.uint32)

    @pytest.fixture
    def array_uri(self):
        uri = self.path("test_read_subarray_array_1d")
        schema = tiledb.ArraySchema(
            domain=tiledb.Domain(
                tiledb.Dim(name="d1", domain=(-1.0, 1.0), tile=2.0, dtype=np.float64)
            ),
            attrs=[tiledb.Attr(name="a1", dtype=np.uint32)],
            sparse=True,
        )
        tiledb.Array.create(uri, schema)
        with tiledb.open(uri, "w") as array:
            array[self.data_dim1] = self.data_attr1
        return uri

    def test_read_full_array(self, array_uri):
        with tiledb.open(array_uri, "r") as array:
            subarray = tiledb.Subarray(array)
            subarray.add_dim_range(0, (-1.0, 1.0))
            result = array.read_subarray(subarray)

        expected = OrderedDict([("d1", self.data_dim1), ("a1", self.data_attr1)])

        assert_dict_arrays_equal(result, expected, False)

    def test_empty_result(self, array_uri):
        with tiledb.open(array_uri, "r") as array:
            subarray = tiledb.Subarray(array)
            subarray.add_dim_range(0, (-0.9, -0.89))
            result = array.read_subarray(subarray)

        expected = OrderedDict(
            [
                ("d1", np.array([], dtype=np.float64)),
                ("a1", np.array([], dtype=np.uint32)),
            ]
        )
        assert_dict_arrays_equal(result, expected, True)



import io
from typing import List, Optional, Sequence, Union, overload

import numpy as np

import tiledb.libtiledb as lt

from .ctx import Ctx, CtxMixin
from .datatypes import DataType


class Filter(CtxMixin, lt.Filter):
    """Base class for all TileDB filters."""

    options: Sequence[lt.FilterOption] = ()

    def __init__(self, type: lt.FilterOption, ctx: Optional[Ctx] = None):
        super().__init__(ctx, type)

    def __repr__(self) -> str:
        # use safe repr if pybind11 constructor failed
        if self._ctx is None:
            return object.__repr__(self)

        output = io.StringIO()
        output.write(f"{type(self).__name__}(")
        if hasattr(self, "_attrs_"):
            attr_output = []
            for f in self._attrs_():
                a = getattr(self, f)
                attr_output.append(f"{f}={a}")
            output.write(",".join(attr_output))
        output.write(")")
        return output.getvalue()

    def _repr_html_(self) -> str:
        output = io.StringIO()

        opt = list(self._attrs_().keys())[0] if self._attrs_() else ""
        val = getattr(self, opt) if self._attrs_() else ""

        output.write("<section>\n")
        output.write("<table>\n")
        output.write("<tr>\n")
        output.write(f"<td>{type(self).__name__}</td>\n")
        output.write(f"<td>{opt}</td>")
        output.write(f"<td>{val}</td>")
        output.write("</tr>\n")
        output.write("</table>\n")
        output.write("</section>\n")

        return output.getvalue()

    def __eq__(self, other: "Filter"):
        if other.__class__ is not self.__class__:
            return False
        for f in self._attrs_():
            left = getattr(self, f)
            right = getattr(other, f)
            if left != right:
                return False
        return True

    @property
    def filter_name(self):
        return self._type.name


class CompressionFilter(Filter):
    """
    Base class for filters performing compression.

    All compression filters support a compression level option, although
    some (such as RLE) ignore it.

    **Example:**

    >>> import tiledb, numpy as np, tempfile
    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     dom = tiledb.Domain(tiledb.Dim(domain=(0, 9), tile=2, dtype=np.uint64))
    ...     a1 = tiledb.Attr(name="a1", dtype=np.int64,
    ...                      filters=tiledb.FilterList([tiledb.CompressionFilter(level=10)]))
    ...     schema = tiledb.ArraySchema(domain=dom, attrs=(a1,))
    ...     tiledb.DenseArray.create(tmp + "/array", schema)

    """

    options = (lt.FilterOption.COMPRESSION_LEVEL,)

    def __init__(self, type: lt.FilterType, level: int = -1, ctx: Optional[Ctx] = None):
        if not isinstance(level, int):
            raise ValueError("`level` argument must be a int")

        super().__init__(type, ctx)
        self._set_option(self._ctx, lt.FilterOption.COMPRESSION_LEVEL, level)

    @property
    def level(self):
        return self._get_option(self._ctx, lt.FilterOption.COMPRESSION_LEVEL)


class NoOpFilter(Filter):
    """A filter that does nothing."""

    def __init__(self, ctx: Optional[Ctx] = None):
        super().__init__(lt.FilterType.NONE, ctx)

    def _attrs_(self):
        return {}


class GzipFilter(CompressionFilter):
    """
    Filter that compresses using gzip.

    :param ctx: TileDB Ctx
    :type ctx: tiledb.Ctx
    :param level: -1 (default) sets the compressor level to the default level as specified in TileDB core. Otherwise, sets the compressor level to the given value.
    :type level: int

    **Example:**

    >>> import tiledb, numpy as np, tempfile
    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     dom = tiledb.Domain(tiledb.Dim(domain=(0, 9), tile=2, dtype=np.uint64))
    ...     a1 = tiledb.Attr(name="a1", dtype=np.int64,
    ...                      filters=tiledb.FilterList([tiledb.GzipFilter()]))
    ...     schema = tiledb.ArraySchema(domain=dom, attrs=(a1,))
    ...     tiledb.DenseArray.create(tmp + "/array", schema)

    """

    def __init__(self, level: int = -1, ctx: Optional[Ctx] = None):
        if not isinstance(level, int):
            raise ValueError("`level` argument must be a int")

        super().__init__(lt.FilterType.GZIP, level, ctx)

    def _attrs_(self):
        return {"level": self.level}


class ZstdFilter(CompressionFilter):
    """
    Filter that compresses using zstd.

    :param ctx: TileDB Ctx
    :type ctx: tiledb.Ctx
    :param level: -1 (default) sets the compressor level to the default level as specified in TileDB core. Otherwise, sets the compressor level to the given value.
    :type level: int

    **Example:**

    >>> import tiledb, numpy as np, tempfile
    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     dom = tiledb.Domain(tiledb.Dim(domain=(0, 9), tile=2, dtype=np.uint64))
    ...     a1 = tiledb.Attr(name="a1", dtype=np.int64,
    ...                      filters=tiledb.FilterList([tiledb.ZstdFilter()]))
    ...     schema = tiledb.ArraySchema(domain=dom, attrs=(a1,))
    ...     tiledb.DenseArray.create(tmp + "/array", schema)

    """

    def __init__(self, level: int = -1, ctx: Optional[Ctx] = None):
        if not isinstance(level, int):
            raise ValueError("`level` argument must be a int")

        super().__init__(lt.FilterType.ZSTD, level, ctx)

    def _attrs_(self):
        return {"level": self.level}


class LZ4Filter(CompressionFilter):
    """
    Filter that compresses using lz4.

    :param ctx: TileDB Ctx
    :type ctx: tiledb.Ctx
    :param level: -1 (default) sets the compressor level to the default level as specified in TileDB core. Otherwise, sets the compressor level to the given value.
    :type level: int

    **Example:**

    >>> import tiledb, numpy as np, tempfile
    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     dom = tiledb.Domain(tiledb.Dim(domain=(0, 9), tile=2, dtype=np.uint64))
    ...     a1 = tiledb.Attr(name="a1", dtype=np.int64,
    ...                      filters=tiledb.FilterList([tiledb.LZ4Filter()]))
    ...     schema = tiledb.ArraySchema(domain=dom, attrs=(a1,))
    ...     tiledb.DenseArray.create(tmp + "/array", schema)

    """

    def __init__(self, level: int = -1, ctx: Optional[Ctx] = None):
        if not isinstance(level, int):
            raise ValueError("`level` argument must be a int")

        super().__init__(lt.FilterType.LZ4, level, ctx)

    def _attrs_(self):
        return {"level": self.level}


class Bzip2Filter(CompressionFilter):
    """
    Filter that compresses using bzip2.

    :param level: -1 (default) sets the compressor level to the default level as specified in TileDB core. Otherwise, sets the compressor level to the given value.
    :type level: int

    **Example:**

    >>> import tiledb, numpy as np, tempfile
    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     dom = tiledb.Domain(tiledb.Dim(domain=(0, 9), tile=2, dtype=np.uint64))
    ...     a1 = tiledb.Attr(name="a1", dtype=np.int64,
    ...                      filters=tiledb.FilterList([tiledb.Bzip2Filter()]))
    ...     schema = tiledb.ArraySchema(domain=dom, attrs=(a1,))
    ...     tiledb.DenseArray.create(tmp + "/array", schema)

    """

    def __init__(self, level: int = -1, ctx: Optional[Ctx] = None):
        if not isinstance(level, int):
            raise ValueError("`level` argument must be a int")

        super().__init__(lt.FilterType.BZIP2, level, ctx)

    def _attrs_(self):
        return {"level": self.level}


class RleFilter(CompressionFilter):
    """
    Filter that compresses using run-length encoding (RLE).

    :param level: -1 (default) sets the compressor level to the default level as specified in TileDB core. Otherwise, sets the compressor level to the given value.
    :type level: int

    **Example:**

    >>> import tiledb, numpy as np, tempfile
    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     dom = tiledb.Domain(tiledb.Dim(domain=(0, 9), tile=2, dtype=np.uint64))
    ...     a1 = tiledb.Attr(name="a1", dtype=np.int64,
    ...                      filters=tiledb.FilterList([tiledb.RleFilter()]))
    ...     schema = tiledb.ArraySchema(domain=dom, attrs=(a1,))
    ...     tiledb.DenseArray.create(tmp + "/array", schema)

    """

    def __init__(self, level: int = -1, ctx: Optional[Ctx] = None):
        if not isinstance(level, int):
            raise ValueError("`level` argument must be a int")

        super().__init__(lt.FilterType.RLE, level, ctx)

    def _attrs_(self):
        return {}


class DeltaFilter(CompressionFilter):
    """
    Filter that compresses using run-length encoding (RLE).

    :param level: -1 (default) sets the compressor level to the default level as specified in TileDB core. Otherwise, sets the compressor level to the given value.
    :type level: int
    :param reinterp_dtype: (optional) sets the compressor to compress the data treating
    as the new datatype.
    :type reinterp_dtype: numpy, lt.DataType
    **Example:**

    >>> import tiledb, numpy as np, tempfile
    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     dom = tiledb.Domain(tiledb.Dim(domain=(0, 9), tile=2, dtype=np.uint64))
    ...     a1 = tiledb.Attr(name="a1", dtype=np.int64,
    ...                      filters=tiledb.FilterList([tiledb.DeltaFilter()]))
    ...     schema = tiledb.ArraySchema(domain=dom, attrs=(a1,))
    ...     tiledb.DenseArray.create(tmp + "/array", schema)

    """

    options = (
        lt.FilterOption.COMPRESSION_LEVEL,
        lt.FilterOption.COMPRESSION_REINTERPRET_DATATYPE,
    )

    def __init__(
        self,
        level: int = -1,
        reinterp_dtype: Optional[Union[np.dtype, lt.DataType]] = None,
        ctx: Optional[Ctx] = None,
    ):
        if not isinstance(level, int):
            raise ValueError("`level` argument must be a int")

        super().__init__(lt.FilterType.DELTA, level, ctx)

        if reinterp_dtype is not None:
            if isinstance(reinterp_dtype, lt.DataType):
                dtype = reinterp_dtype
            else:
                dtype = DataType.from_numpy(reinterp_dtype).tiledb_type
            self._set_option(
                self._ctx, lt.FilterOption.COMPRESSION_REINTERPRET_DATATYPE, dtype
            )

    def _attrs_(self):
        return {"reinterp_dtype": self.reinterp_dtype}

    @property
    def reinterp_dtype(self):
        tiledb_dtype = self._get_option(
            self._ctx, lt.FilterOption.COMPRESSION_REINTERPRET_DATATYPE
        )
        if tiledb_dtype == lt.DataType.ANY:
            return None
        dtype = DataType.from_tiledb(tiledb_dtype)
        return dtype.np_dtype


class DoubleDeltaFilter(CompressionFilter):
    """
    Filter that performs double-delta encoding.

    :param level: -1 (default) sets the compressor level to the default level as specified in TileDB core. Otherwise, sets the compressor level to the given value.
    :type level: int
    :param reinterp_dtype: (optional) sets the compressor to compress the data treating as the new datatype.

    **Example:**

    >>> import tiledb, numpy as np, tempfile
    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     dom = tiledb.Domain(tiledb.Dim(domain=(0, 9), tile=2, dtype=np.uint64))
    ...     a1 = tiledb.Attr(name="a1", dtype=np.int64,
    ...                      filters=tiledb.FilterList([tiledb.DoubleDeltaFilter()]))
    ...     schema = tiledb.ArraySchema(domain=dom, attrs=(a1,))
    ...     tiledb.DenseArray.create(tmp + "/array", schema)

    """

    options = (
        lt.FilterOption.COMPRESSION_LEVEL,
        lt.FilterOption.COMPRESSION_REINTERPRET_DATATYPE,
    )

    def __init__(
        self,
        level: int = -1,
        reinterp_dtype: Optional[Union[np.dtype, lt.DataType]] = None,
        ctx: Optional[Ctx] = None,
    ):
        if not isinstance(level, int):
            raise ValueError("`level` argument must be a int")

        super().__init__(lt.FilterType.DOUBLE_DELTA, level, ctx)

        if reinterp_dtype is not None:
            if isinstance(reinterp_dtype, lt.DataType):
                dtype = reinterp_dtype
            else:
                dtype = DataType.from_numpy(reinterp_dtype).tiledb_type
            self._set_option(
                self._ctx, lt.FilterOption.COMPRESSION_REINTERPRET_DATATYPE, dtype
            )

    def _attrs_(self):
        return {"reinterp_dtype": self.reinterp_dtype}

    @property
    def reinterp_dtype(self):
        tiledb_dtype = self._get_option(
            self._ctx, lt.FilterOption.COMPRESSION_REINTERPRET_DATATYPE
        )
        if tiledb_dtype == lt.DataType.ANY:
            return None
        dtype = DataType.from_tiledb(tiledb_dtype)
        return dtype.np_dtype


class DictionaryFilter(CompressionFilter):
    """
    Filter that performs dictionary encoding.

    :param level: -1 (default) sets the compressor level to the default level as specified in TileDB core. Otherwise, sets the compressor level to the given value.
    :type level: int

    **Example:**

    >>> import tiledb, numpy as np, tempfile
    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     dom = tiledb.Domain(tiledb.Dim(domain=(0, 9), tile=2, dtype=np.uint64))
    ...     a1 = tiledb.Attr(name="a1", dtype=np.int64,
    ...                      filters=tiledb.FilterList([tiledb.DictionaryFilter()]))
    ...     schema = tiledb.ArraySchema(domain=dom, attrs=(a1,))
    ...     tiledb.DenseArray.create(tmp + "/array", schema)

    """

    def __init__(self, level: int = -1, ctx: Optional[Ctx] = None):
        if not isinstance(level, int):
            raise ValueError("`level` argument must be a int")

        super().__init__(lt.FilterType.DICTIONARY, level, ctx)

    def _attrs_(self):
        return {}


class BitShuffleFilter(Filter):
    """
    Filter that performs a bit shuffle transformation.

    **Example:**

    >>> import tiledb, numpy as np, tempfile
    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     dom = tiledb.Domain(tiledb.Dim(domain=(0, 9), tile=2, dtype=np.uint64))
    ...     a1 = tiledb.Attr(name="a1", dtype=np.int64,
    ...                      filters=tiledb.FilterList([tiledb.BitShuffleFilter()]))
    ...     schema = tiledb.ArraySchema(domain=dom, attrs=(a1,))
    ...     tiledb.DenseArray.create(tmp + "/array", schema)

    """

    def __init__(self, ctx: Optional[Ctx] = None):
        super().__init__(lt.FilterType.BITSHUFFLE, ctx)

    def _attrs_(self):
        return {}


class ByteShuffleFilter(Filter):
    """
    Filter that performs a byte shuffle transformation.

    **Example:**

    >>> import tiledb, numpy as np, tempfile
    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     dom = tiledb.Domain(tiledb.Dim(domain=(0, 9), tile=2, dtype=np.uint64))
    ...     a1 = tiledb.Attr(name="a1", dtype=np.int64,
    ...                      filters=tiledb.FilterList([tiledb.ByteShuffleFilter()]))
    ...     schema = tiledb.ArraySchema(domain=dom, attrs=(a1,))
    ...     tiledb.DenseArray.create(tmp + "/array", schema)

    """

    def __init__(self, ctx: Optional[Ctx] = None):
        super().__init__(lt.FilterType.BYTESHUFFLE, ctx)

    def _attrs_(self):
        return {}


class BitWidthReductionFilter(Filter):
    """Filter that performs bit-width reduction.

     :param ctx: A TileDB Context
     :type ctx: tiledb.Ctx
     :param window: -1 (default) sets the max window size for the filter to the default window size as specified in TileDB core. Otherwise, sets the compressor level to the given value.
     :type window: int

    **Example:**

    >>> import tiledb, numpy as np, tempfile
    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     dom = tiledb.Domain(tiledb.Dim(domain=(0, 9), tile=2, dtype=np.uint64))
    ...     a1 = tiledb.Attr(name="a1", dtype=np.int64,
    ...                      filters=tiledb.FilterList([tiledb.BitWidthReductionFilter()]))
    ...     schema = tiledb.ArraySchema(domain=dom, attrs=(a1,))
    ...     tiledb.DenseArray.create(tmp + "/array", schema)

    """

    options = (lt.FilterOption.BIT_WIDTH_MAX_WINDOW,)

    def __init__(self, window: int = -1, ctx: Optional[Ctx] = None):
        if not isinstance(window, int):
            raise ValueError("`window` argument must be a int")

        super().__init__(lt.FilterType.BIT_WIDTH_REDUCTION, ctx)

        if window != -1:
            self._set_option(self._ctx, lt.FilterOption.BIT_WIDTH_MAX_WINDOW, window)

    def _attrs_(self):
        return {"window": self.window}

    @property
    def window(self):
        return self._get_option(self._ctx, lt.FilterOption.BIT_WIDTH_MAX_WINDOW)


class PositiveDeltaFilter(Filter):
    """
    Filter that performs positive-delta encoding.

    :param ctx: A TileDB Context
    :type ctx: tiledb.Ctx
    :param window: -1 (default) sets the max window size for the filter to the default window size as specified in TileDB core. Otherwise, sets the compressor level to the given value.
    :type window: int

    **Example:**

    >>> import tiledb, numpy as np, tempfile
    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     dom = tiledb.Domain(tiledb.Dim(domain=(0, 9), tile=2, dtype=np.uint64))
    ...     a1 = tiledb.Attr(name="a1", dtype=np.int64,
    ...                      filters=tiledb.FilterList([tiledb.PositiveDeltaFilter()]))
    ...     schema = tiledb.ArraySchema(domain=dom, attrs=(a1,))
    ...     tiledb.DenseArray.create(tmp + "/array", schema)

    """

    options = (lt.FilterOption.POSITIVE_DELTA_MAX_WINDOW,)

    def __init__(self, window: int = -1, ctx: Optional[Ctx] = None):
        if not isinstance(window, int):
            raise ValueError("`window` argument must be a int")

        super().__init__(lt.FilterType.POSITIVE_DELTA, ctx)
        if window != -1:
            self._set_option(
                self._ctx, lt.FilterOption.POSITIVE_DELTA_MAX_WINDOW, window
            )

    def _attrs_(self):
        return {"window": self.window}

    @property
    def window(self):
        return self._get_option(self._ctx, lt.FilterOption.POSITIVE_DELTA_MAX_WINDOW)


class ChecksumMD5Filter(Filter):
    """
    MD5 checksum filter.

    **Example:**

    >>> import tiledb, numpy as np, tempfile
    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     dom = tiledb.Domain(tiledb.Dim(domain=(0, 9), tile=2, dtype=np.uint64))
    ...     a1 = tiledb.Attr(name="a1", dtype=np.int64,
    ...                      filters=tiledb.FilterList([tiledb.ChecksumMD5Filter()]))
    ...     schema = tiledb.ArraySchema(domain=dom, attrs=(a1,))
    ...     tiledb.DenseArray.create(tmp + "/array", schema)

    """

    def __init__(self, ctx: Optional[Ctx] = None):
        super().__init__(lt.FilterType.CHECKSUM_MD5, ctx)

    def _attrs_(self):
        return {}


class ChecksumSHA256Filter(Filter):
    """
    SHA256 checksum filter.

    **Example:**

    >>> import tiledb, numpy as np, tempfile
    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     dom = tiledb.Domain(tiledb.Dim(domain=(0, 9), tile=2, dtype=np.uint64))
    ...     a1 = tiledb.Attr(name="a1", dtype=np.int64,
    ...                      filters=tiledb.FilterList([tiledb.ChecksumSHA256Filter()]))
    ...     schema = tiledb.ArraySchema(domain=dom, attrs=(a1,))
    ...     tiledb.DenseArray.create(tmp + "/array", schema)

    """

    def __init__(self, ctx: Optional[Ctx] = None):
        super().__init__(lt.FilterType.CHECKSUM_SHA256, ctx)

    def _attrs_(self):
        return {}


class FloatScaleFilter(Filter):
    """
    Filter that stores floats as integers in a reduced representation via scaling.
    The reduced storage space is in lieu of some precision loss. The float scaling
    filter takes three parameters: the factor, the offset, and the bytewidth.
    On write, the float scaling filter applies the factor (scaling factor) and offset,
    and stores the value of round((raw_float - offset) / factor) as an
    integer with the specified NumPy dtype.
    On read, the float scaling filter will reverse the factor and offset,
    and returns the floating point data, with a potential loss of precision.
    :param factor: the scaling factor used to translate the data
    :type factor: float
    :param offset: the offset value used to translate the data
    :type offset: float
    :param bytewidth: values may be stored as integers of bytewidth 1, 2, 4, or 8
    :type np.integer:
    :param ctx: A TileDB Context
    :type ctx: tiledb.Ctx
    **Example:**
    >>> import tiledb, numpy as np, tempfile
    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     dom = tiledb.Domain(tiledb.Dim(domain=(0, 9), tile=2, dtype=np.uint64))
    ...     a1 = tiledb.Attr(name="a1", dtype=np.int64,
    ...                      filters=tiledb.FilterList([tiledb.FloatScaleFilter(1, 0)]))
    ...     schema = tiledb.ArraySchema(domain=dom, attrs=(a1,))
    ...     tiledb.DenseArray.create(tmp + "/array", schema)
    """

    options = (
        lt.FilterOption.SCALE_FLOAT_FACTOR,
        lt.FilterOption.SCALE_FLOAT_OFFSET,
        lt.FilterOption.SCALE_FLOAT_BYTEWIDTH,
    )

    def __init__(
        self,
        factor: float = None,
        offset: float = None,
        bytewidth: int = None,
        ctx: Optional[Ctx] = None,
    ):
        self._factor = factor
        self._offset = offset
        self._bytewidth = bytewidth
        super().__init__(lt.FilterType.SCALE_FLOAT, ctx)

        if factor:
            self._set_option(
                self._ctx, lt.FilterOption.SCALE_FLOAT_FACTOR, float(factor)
            )

        if offset is not None:
            # 0 evals to false so we weren't actually setting here
            # if offset==0 :upside-down-face:
            self._set_option(
                self._ctx, lt.FilterOption.SCALE_FLOAT_OFFSET, float(offset)
            )

        if bytewidth:
            self._set_option(
                self._ctx, lt.FilterOption.SCALE_FLOAT_BYTEWIDTH, bytewidth
            )

    def dump(self):
        print(self._dump(), "\n")

    def _attrs_(self):
        return {
            "factor": self._factor,
            "offset": self._offset,
            "bytewidth": self._bytewidth,
        }

    @property
    def factor(self):
        return self._get_option(self._ctx, lt.FilterOption.SCALE_FLOAT_FACTOR)

    @property
    def offset(self):
        return self._get_option(self._ctx, lt.FilterOption.SCALE_FLOAT_OFFSET)

    @property
    def bytewidth(self):
        return self._get_option(self._ctx, lt.FilterOption.SCALE_FLOAT_BYTEWIDTH)


class XORFilter(Filter):
    """
    XOR  filter.

    **Example:**

    >>> import tiledb, numpy as np, tempfile
    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     dom = tiledb.Domain(tiledb.Dim(domain=(0, 9), tile=2, dtype=np.uint64))
    ...     a1 = tiledb.Attr(name="a1", dtype=np.int64,
    ...                      filters=tiledb.FilterList([tiledb.XORFilter()]))
    ...     schema = tiledb.ArraySchema(domain=dom, attrs=(a1,))
    ...     tiledb.DenseArray.create(tmp + "/array", schema)

    """

    def __init__(self, ctx: Optional[Ctx] = None):
        super().__init__(lt.FilterType.XOR, ctx)

    def _attrs_(self):
        return {}


class WebpFilter(Filter):
    """
    The WebP filter provides three options: quality, format, and lossless

    The quality option is used as quality_factor setting for WebP lossy
    compression and expects a float value in the range of 0.0f - 100.0f
    Quality of 0 corresponds to low quality and small output sizes, whereas 100
    is the highest quality and largest output size.

    The format option is used to define colorspace format of image data and
    expects an enum of TILEDB_WEBP_RGB, TILEDB_WEBP_BGR, TILEDB_WEBP_RGBA, or
    TILEDB_WEBP_BGRA.

    The lossless option is used to enable(1) or disable(0) lossless compression.
    With this option enabled, the quality setting will be ignored.

    On write this filter takes raw colorspace values (RGB, RBGA, etc) and encodes
    into WebP format before writing data to the array.

    On read, this filter decodes WebP data and returns raw colorspace values to
    the caller.

    This filter expects the array to provide two dimensions for Y, X pixel position.
    Dimensions may be defined with any name, but Y, X should be at dimension index 0, 1 respectively.
    Dimensions can be any two matching integral types, such as `{uint64_t, uint64_t}` or `{int64_t, int64_t}`.

    The WebP filter supports only the `uint8_t` type for attributes.

    :param quality: quality_factor setting for lossy WebP compression
    :type quality: float in range [0.0, 100.0]
    :param input_format: The input colorspace format of the image
    :type input_format: np.uint8 corresponding to one of TILEDB_WEBP_{RGB, BGR, RGBA, BGRA}
    :param lossless: Enable (1) or disable (0) lossless image compression
    :type lossless: np.uint8
    :param ctx: A TileDB Context
    :type ctx: tiledb.Ctx

    **Example:**

    >>> import tiledb, numpy as np, tempfile
    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     # Using RGB colorspace format
    ...     pixel_depth = 3  # For RGBA / BGRA pixel_depth is 4
    ...     dims = (tiledb.Dim(name='Y',
    ...                    domain=(1, img_height),
    ...                    dtype=np.uint8,
    ...                    tile=img_height / 2,),
    ...             tiledb.Dim(name='X',
    ...                    domain=(1, img_width * pixel_depth),
    ...                    dtype=np.uint8,
    ...                    tile=(img_width / 2) * pixel_depth,))
    ...     dom = tiledb.Domain(*dims)
    ...     rgb = tiledb.Attr(name="rgb", dtype=np.uint8,
    ...                      filters=tiledb.FilterList([tiledb.WebpFilter(input_format=1, quality=100.0, lossless=1)]))
    ...     schema = tiledb.ArraySchema(domain=dom, attrs=(rgb,))
    ...     tiledb.DenseArray.create(tmp + "/array", schema)

    """

    options = (
        lt.FilterOption.WEBP_INPUT_FORMAT,
        lt.FilterOption.WEBP_QUALITY,
        lt.FilterOption.WEBP_LOSSLESS,
    )

    # Expose WebP enums at the top level
    WebpInputFormat = lt.WebpInputFormat

    def __init__(
        self,
        input_format: lt.WebpInputFormat = None,
        quality: float = None,
        lossless: bool = None,
        ctx: Optional[Ctx] = None,
    ):
        self._input_format = input_format
        self._quality = quality
        self._lossless = lossless
        super().__init__(lt.FilterType.WEBP, ctx)

        if input_format is not None:
            self._set_option(
                self._ctx,
                lt.FilterOption.WEBP_INPUT_FORMAT,
                lt.WebpInputFormat(input_format),
            )

        if quality is not None:
            self._set_option(self._ctx, lt.FilterOption.WEBP_QUALITY, float(quality))

        if lossless is not None:
            self._set_option(self._ctx, lt.FilterOption.WEBP_LOSSLESS, lossless)

    def _attrs_(self):
        return {
            "input_format": self._input_format,
            "quality": self._quality,
            "lossless": self._lossless,
        }

    @property
    def input_format(self):
        return lt.WebpInputFormat(
            self._get_option(self._ctx, lt.FilterOption.WEBP_INPUT_FORMAT)
        )

    @property
    def quality(self):
        return self._get_option(self._ctx, lt.FilterOption.WEBP_QUALITY)

    @property
    def lossless(self):
        return self._get_option(self._ctx, lt.FilterOption.WEBP_LOSSLESS)


#
class FilterList(CtxMixin, lt.FilterList):
    """
    An ordered list of Filter objects for filtering TileDB data.

    FilterLists contain zero or more Filters, used for filtering attribute data, the array coordinate data, etc.

    :param ctx: A TileDB context
    :type ctx: tiledb.Ctx
    :param filters: An iterable of Filter objects to add.
    :param chunksize: (default None) chunk size used by the filter list in bytes
    :type chunksize: int

    **Example:**

    >>> import tiledb, numpy as np, tempfile
    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     dom = tiledb.Domain(tiledb.Dim(domain=(0, 9), tile=2, dtype=np.uint64))
    ...     # Create several filters
    ...     gzip_filter = tiledb.GzipFilter()
    ...     bw_filter = tiledb.BitWidthReductionFilter()
    ...     # Create a filter list that will first perform bit width reduction, then gzip compression.
    ...     filters = tiledb.FilterList([bw_filter, gzip_filter])
    ...     a1 = tiledb.Attr(name="a1", dtype=np.int64, filters=filters)
    ...     # Create a second attribute filtered only by gzip compression.
    ...     a2 = tiledb.Attr(name="a2", dtype=np.int64,
    ...                      filters=tiledb.FilterList([gzip_filter]))
    ...     schema = tiledb.ArraySchema(domain=dom, attrs=(a1, a2))
    ...     tiledb.DenseArray.create(tmp + "/array", schema)

    """

    filter_type_cc_to_python = {
        lt.FilterType.GZIP: GzipFilter,
        lt.FilterType.ZSTD: ZstdFilter,
        lt.FilterType.LZ4: LZ4Filter,
        lt.FilterType.BZIP2: Bzip2Filter,
        lt.FilterType.RLE: RleFilter,
        lt.FilterType.DELTA: DeltaFilter,
        lt.FilterType.DOUBLE_DELTA: DoubleDeltaFilter,
        lt.FilterType.BIT_WIDTH_REDUCTION: BitWidthReductionFilter,
        lt.FilterType.BITSHUFFLE: BitShuffleFilter,
        lt.FilterType.BYTESHUFFLE: ByteShuffleFilter,
        lt.FilterType.POSITIVE_DELTA: PositiveDeltaFilter,
        lt.FilterType.CHECKSUM_MD5: ChecksumMD5Filter,
        lt.FilterType.CHECKSUM_SHA256: ChecksumSHA256Filter,
        lt.FilterType.DICTIONARY: DictionaryFilter,
        lt.FilterType.SCALE_FLOAT: FloatScaleFilter,
        lt.FilterType.XOR: XORFilter,
        lt.FilterType.WEBP: WebpFilter,
        lt.FilterType.NONE: NoOpFilter,
    }

    def __init__(
        self,
        filters: Sequence[Filter] = None,
        chunksize: int = None,
        ctx: Optional[Ctx] = None,
    ):
        super().__init__(ctx)
        if filters is not None:
            for f in filters:
                if not isinstance(f, Filter):
                    raise ValueError(
                        "filters argument must be an iterable of TileDB filter objects"
                    )
                self._add_filter(f)
        if chunksize is not None:
            self._chunksize = chunksize

    @property
    def chunksize(self):
        return self._chunksize

    @overload
    def __getitem__(self, idx: int) -> Filter:
        ...

    @overload
    def __getitem__(self, idx: slice) -> List[Filter]:
        ...

    def __getitem__(self, idx):
        """Gets a copy of the filter in the list at the given index

        :param idx: index into the
        :type idx: int or slice
        :returns: A filter at given index / slice
        :raises IndexError: invalid index
        :raises: :py:exc:`tiledb.TileDBError`

        """
        if not isinstance(idx, (int, slice)):
            raise TypeError(
                "FilterList indices must be integers or slices, not {:s}".format(
                    type(idx).__name__
                )
            )

        if isinstance(idx, int):
            if idx < 0 or idx > (len(self) - 1):
                raise IndexError("FilterList index out of range")
            idx = slice(idx, idx + 1)
        else:
            if (
                not isinstance(idx.start, int)
                or not isinstance(idx.stop, int)
                or not isinstance(idx.step, int)
            ):
                raise IndexError("FilterList slice indices must be integers or None")

        filters = []
        (start, stop, step) = idx.indices(len(self))
        for i in range(start, stop, step):
            filters.append(self._filter(i))

        if len(filters) == 1:
            return filters[0]

        return filters

    def __eq__(self, other: Union["FilterList", Sequence[Filter]]) -> bool:
        if other is None:
            return False
        if len(self) != len(other):
            return False
        for i, f in enumerate(self):
            if f != other[i]:
                return False
        return True

    def __len__(self) -> int:
        """
        :rtype: int
        :return: Number of filters in the FilterList

        """
        return self._nfilters()

    def append(self, filter: Filter):
        """
        :param Filter filter: the filter to append into the FilterList
        :raises ValueError: filter argument incorrect type

        """
        if not isinstance(filter, Filter):
            raise ValueError("filter argument must be a TileDB Filter object")
        self._add_filter(filter)

    def __repr__(self) -> str:
        filters = ",\n       ".join(repr(self._filter(i)) for i in range(len(self)))
        return "FilterList([{0!s}])".format(filters)

    def _repr_html_(self) -> str:
        if len(self) == 0:
            return "-"

        output = io.StringIO()
        output.write("<section>\n")
        output.write("<table>\n")
        output.write("<tr>\n")
        output.write("<th>Name</th>\n")
        output.write("<th>Option</th>\n")
        output.write("<th>Level</th>\n")
        output.write("</tr>\n")
        for filter in self:
            opt = list(filter._attrs_().keys())[0] if filter._attrs_() else "-"
            val = getattr(filter, opt) if filter._attrs_() else "-"
            output.write("<tr>\n")
            output.write(f"<td>{type(filter).__name__}</td>\n")
            output.write(f"<td>{opt}</td>")
            output.write(f"<td>{val}</td>")
            output.write("</tr>\n")
        output.write("</table>\n")
        output.write("</section>\n")

        return output.getvalue()

    def _filter(self, i: int) -> Filter:
        fil = super()._filter(i)
        filtype = self.filter_type_cc_to_python[fil._type]
        ctx = self._ctx
        return filtype(*(fil._get_option(ctx, opt) for opt in filtype.options), ctx=ctx)



import re

import numpy as np
import pytest
from numpy.testing import assert_array_equal

import tiledb

from .common import DiskTestCase, has_pandas, has_pyarrow


class EnumerationTest(DiskTestCase):
    @pytest.mark.parametrize(
        "name,data",
        (
            ("int", np.array([0])),
            ("float", np.array([1.0, 2.2, 5.8234, 94.23])),
            ("str", np.array(["abc", "defghi", "jk"])),
            ("utf8", np.array(["abc", "defghi", "jk"], dtype=np.str_)),
            ("ascii", np.array([b"abc", b"defghi", b"jk"], dtype=np.bytes_)),
        ),
    )
    @pytest.mark.parametrize("ordered", [True, False])
    def test_enumeration_basic(self, name, ordered, data):
        enmr = tiledb.Enumeration(name, ordered, data)

        assert enmr.name == name
        assert enmr.ordered == ordered
        assert_array_equal(enmr.values(), data)
        if name in ("str", "utf8", "ascii"):
            assert enmr.cell_val_num == tiledb.libtiledb.TILEDB_VAR_NUM()
            assert enmr.dtype.kind == data.dtype.kind
        else:
            assert enmr.cell_val_num == 1
            assert enmr.dtype.kind == data.dtype.kind

    def test_attribute_enumeration(self):
        attr = tiledb.Attr()
        attr.enum = "enum"
        assert attr.enum == "enum"

    def test_enumeration_repr(self):
        """Doesn't check exact string, just makes sure each component is matched, in case order is changed in the future."""
        enmr = tiledb.Enumeration("e", False, [1, 2, 3])
        # Get its string representation
        repr_str = repr(enmr)

        # Define patterns to match each component in the representation
        patterns = {
            "Enumeration": r"Enumeration",
            "name": r"name='e'",
            # use regex because it is depending on platform
            "dtype": r"dtype=int\d+",
            "dtype_name": r"dtype_name='int\d+'",
            "cell_val_num": r"cell_val_num=1",
            "ordered": r"ordered=False",
            "values": r"values=\[1, 2, 3\]",
        }

        # Check that each pattern is found in the representation string
        for key, pattern in patterns.items():
            assert re.search(pattern, repr_str), f"{key} not found or incorrect in repr"

    def test_array_schema_enumeration(self):
        uri = self.path("test_array_schema_enumeration")
        dom = tiledb.Domain(tiledb.Dim(domain=(1, 8), tile=1))
        enum1 = tiledb.Enumeration("enmr1", False, np.arange(3) * 10)
        enum2 = tiledb.Enumeration("enmr2", False, ["a", "bb", "ccc"])
        attr1 = tiledb.Attr("attr1", dtype=np.int32, enum_label="enmr1")
        attr2 = tiledb.Attr("attr2", dtype=np.int32, enum_label="enmr2")
        attr3 = tiledb.Attr("attr3", dtype=np.int32)
        schema = tiledb.ArraySchema(
            domain=dom, attrs=(attr1, attr2, attr3), enums=(enum1, enum2)
        )
        tiledb.Array.create(uri, schema)

        data1 = np.random.randint(0, 3, 8)
        data2 = np.random.randint(0, 3, 8)
        data3 = np.random.randint(0, 3, 8)

        with tiledb.open(uri, "w") as A:
            A[:] = {"attr1": data1, "attr2": data2, "attr3": data3}

        with tiledb.open(uri, "r") as A:
            assert A.enum("enmr1") == enum1
            assert attr1.enum_label == "enmr1"
            assert A.attr("attr1").enum_label == "enmr1"

            assert A.enum("enmr2") == enum2
            assert attr2.enum_label == "enmr2"
            assert A.attr("attr2").enum_label == "enmr2"

            with self.assertRaises(tiledb.TileDBError) as excinfo:
                assert A.enum("enmr3") == []
            assert (
                "ArraySchema: Unable to check if unknown enumeration is loaded. No enumeration named 'enmr3'."
                == str(excinfo.value)
            )
            assert attr3.enum_label is None
            assert A.attr("attr3").enum_label is None

            if has_pandas():
                assert_array_equal(A.df[:]["attr1"].cat.codes, data1)
                assert_array_equal(A.df[:]["attr2"].cat.codes, data2)

                assert_array_equal(A.df[:]["attr1"], A.multi_index[:]["attr1"])
                assert_array_equal(A.df[:]["attr2"], A.multi_index[:]["attr2"])

                assert_array_equal(A.df[:]["attr1"], A[:]["attr1"])
                assert_array_equal(A.df[:]["attr2"], A[:]["attr2"])

    @pytest.mark.skipif(
        not has_pyarrow() or not has_pandas(),
        reason="pyarrow>=1.0 and/or pandas>=1.0,<3.0 not installed",
    )
    @pytest.mark.parametrize("sparse", [True, False])
    @pytest.mark.parametrize("pass_df", [True, False])
    def test_array_schema_enumeration_nullable(self, sparse, pass_df):
        import pyarrow as pa

        uri = self.path("test_array_schema_enumeration_nullable")
        enmr = tiledb.Enumeration("e", False, ["alpha", "beta", "gamma"])
        dom = tiledb.Domain(tiledb.Dim("d", domain=(1, 5), dtype="int64"))
        att = tiledb.Attr("a", dtype="int8", nullable=True, enum_label="e")
        schema = tiledb.ArraySchema(
            domain=dom, attrs=[att], enums=[enmr], sparse=sparse
        )
        tiledb.Array.create(uri, schema)

        with tiledb.open(uri, "w") as A:
            dims = pa.array([1, 2, 3, 4, 5])
            data = pa.array([1.0, 2.0, None, 0, 1.0])
            if pass_df:
                dims = dims.to_pandas()
                data = data.to_pandas()

            if sparse:
                A[dims] = data
            else:
                A[:] = data

        with tiledb.open(uri, "r") as A:
            expected_validity = [False, False, True, False, False]
            assert_array_equal(A[:]["a"].mask, expected_validity)
            assert_array_equal(A.df[:]["a"].isna(), expected_validity)
            assert_array_equal(A.query(attrs=["a"])[:]["a"].mask, expected_validity)

    @pytest.mark.parametrize(
        "dtype, values",
        [
            (np.int8, np.array([1, 2, 3], np.int8)),
            (np.uint8, np.array([1, 2, 3], np.uint8)),
            (np.int16, np.array([1, 2, 3], np.int16)),
            (np.uint16, np.array([1, 2, 3], np.uint16)),
            (np.int32, np.array([1, 2, 3], np.int32)),
            (np.uint32, np.array([1, 2, 3], np.uint32)),
            (np.int64, np.array([1, 2, 3], np.int64)),
            (np.uint64, np.array([1, 2, 3], np.uint64)),
            (np.dtype("S"), np.array(["a", "b", "c"], np.dtype("S"))),
            (np.dtype("U"), np.array(["a", "b", "c"], np.dtype("U"))),
        ],
    )
    def test_enum_dtypes(self, dtype, values):
        # create empty
        enmr = tiledb.Enumeration("e", False, dtype=dtype)
        if dtype in (np.dtype("S"), np.dtype("U")):
            assert enmr.dtype.kind == enmr.values().dtype.kind == dtype.kind
        else:
            assert enmr.dtype == enmr.values().dtype == dtype
            assert_array_equal(enmr.values(), [])

        # then extend with values
        enmr = enmr.extend(values)
        if dtype in (np.dtype("S"), np.dtype("U")):
            assert enmr.dtype.kind == enmr.values().dtype.kind == dtype.kind
        else:
            assert enmr.dtype == enmr.values().dtype == dtype
            assert_array_equal(enmr.values(), values)

        # create with values
        enmr = tiledb.Enumeration("e", False, values=values)
        if dtype in (np.dtype("S"), np.dtype("U")):
            assert enmr.dtype.kind == enmr.values().dtype.kind == dtype.kind
        else:
            assert enmr.dtype == enmr.values().dtype == dtype
            assert_array_equal(enmr.values(), values)

    @pytest.mark.skipif(not has_pandas(), reason="pandas>=1.0,<3.0 not installed")
    def test_from_pandas_dtype_mismatch(self):
        import pandas as pd

        schema = tiledb.ArraySchema(
            enums=[
                tiledb.Enumeration(name="enum1", values=["a", "b", "c"], ordered=False)
            ],
            domain=tiledb.Domain(
                tiledb.Dim(name="dim1", dtype=np.int32, domain=(0, 1))
            ),
            attrs=[tiledb.Attr(name="attr1", dtype=np.int32, enum_label="enum1")],
            sparse=True,
        )

        # Pandas category's categories matches the TileDB enumeration's values
        df1 = pd.DataFrame(data={"dim1": [0, 1], "attr1": ["b", "c"]})
        df1["attr1"] = pd.Categorical(values=df1.attr1, categories=["a", "b", "c"])

        array_path = self.path("arr1")
        tiledb.Array.create(array_path, schema)
        tiledb.from_pandas(array_path, df1, schema=schema, mode="append")

        actual_values = tiledb.open(array_path).df[:]["attr1"].values.tolist()
        assert actual_values == ["b", "c"]

        # Pandas category's categories does not match the TileDB enumeration's values
        df2 = pd.DataFrame(data={"dim1": [0, 1], "attr1": ["b", "c"]})
        df2["attr1"] = df2["attr1"].astype("category")

        array_path = self.path("arr2")
        tiledb.Array.create(array_path, schema)
        tiledb.from_pandas(array_path, df2, schema=schema, mode="append")

        actual_values = tiledb.open(array_path).df[:]["attr1"].values.tolist()
        assert actual_values == ["b", "c"]



from collections import OrderedDict

import numpy as np
import pytest

import tiledb

from .common import DiskTestCase, assert_array_equal, assert_dict_arrays_equal

SUPPORTED_INTEGER_DTYPES = (
    np.uint8,
    np.uint16,
    np.uint32,
    np.uint64,
    np.int8,
    np.int16,
    np.int32,
    np.int64,
)


SUPPORTED_DATETIME64_RESOLUTION = ("Y", "M", "W", "D", "h", "m", "s", "ms", "us", "ns")


class TestWriteSubarrayDense(DiskTestCase):
    @pytest.mark.parametrize("dim_res", SUPPORTED_DATETIME64_RESOLUTION)
    def test_1d_datetime_full_write(self, dim_res):
        """Create TileDB array, write data, and return the URI."""
        # Create array.
        uri = self.path(f"write_subarray_1d_datetime_{dim_res}")
        start_time = np.datetime64("2000-01-01", dim_res)
        domain = (start_time, start_time + np.timedelta64(99, dim_res))
        schema = tiledb.ArraySchema(
            tiledb.Domain(
                tiledb.Dim(
                    name="d1",
                    domain=domain,
                    tile=100,
                    dtype=np.dtype(f"M8[{dim_res}]"),
                )
            ),
            [tiledb.Attr(name="", dtype=np.float64)],
        )
        tiledb.Array.create(uri, schema)

        # Write data.
        data = np.random.rand(100)
        with tiledb.open(uri, "w") as array:
            subarray = tiledb.Subarray(array)
            subarray.add_dim_range(0, domain)
            array.write_subarray(subarray, data)

        # Check results.
        with tiledb.open(uri, "r") as array:
            result = array[...]
        assert_array_equal(result, data)

    def test_1d_full_write(self):
        # Create array.
        uri = self.path("dense_write_subarray_1d_full_write")
        schema = tiledb.ArraySchema(
            tiledb.Domain(tiledb.Dim(name="d1", domain=(0, 999), tile=1000)),
            [tiledb.Attr(name="", dtype=np.float64)],
        )
        tiledb.Array.create(uri, schema)

        # Write data.
        data = np.random.rand(1000)
        with tiledb.open(uri, "w") as array:
            subarray = tiledb.Subarray(array)
            subarray.add_dim_range(0, (0, 999))
            array.write_subarray(subarray, data)

        # Check results.
        with tiledb.open(uri, "r") as array:
            result = array[...]
        assert_array_equal(result, data)

    def test_1d_partial_write(self):
        # Create array.
        uri = self.path("dense_write_subarray_1d_multiple_partial_writes")
        schema = tiledb.ArraySchema(
            tiledb.Domain(tiledb.Dim(name="d1", domain=(0, 99), tile=100)),
            [tiledb.Attr(name="", dtype=np.float32)],
        )
        tiledb.Array.create(uri, schema)

        # Write data.
        data = np.random.rand(10).astype(np.float32)
        with tiledb.open(uri, "w") as array:
            subarray = tiledb.Subarray(array)
            subarray.add_dim_range(0, (10, 19))
            array.write_subarray(subarray, data)

        # Check results.
        with tiledb.open(uri, "r") as array:
            result = array[10:20]
        assert_array_equal(result, data)

    def test_multidim_set_all_ranges(self):
        # Create array.
        uri = self.path("dense_write_subarray_multidim_set_all_ranges")
        schema = tiledb.ArraySchema(
            tiledb.Domain(
                tiledb.Dim(name="d1", domain=(0, 99), tile=100),
                tiledb.Dim(name="d2", domain=(0, 99), tile=100),
                tiledb.Dim(name="d3", domain=(0, 99), tile=100),
            ),
            [
                tiledb.Attr(name="a1", dtype=np.float64),
                tiledb.Attr(name="a2", dtype=np.float64),
            ],
        )
        tiledb.Array.create(uri, schema)

        # Write data.
        data = OrderedDict(
            [
                ("a1", np.random.rand(1000).reshape((10, 10, 10))),
                ("a2", np.random.rand(1000).reshape((10, 10, 10))),
            ]
        )

        with tiledb.open(uri, "w") as array:
            subarray = tiledb.Subarray(array)
            subarray.add_dim_range(0, (0, 9))
            subarray.add_dim_range(1, (10, 19))
            subarray.add_dim_range(2, (20, 29))
            array.write_subarray(subarray, data)

        # Check results.
        with tiledb.open(uri, "r") as array:
            nonempty = array.nonempty_domain()
            assert nonempty == ((0, 9), (10, 19), (20, 29))
            result = array[0:10, 10:20, 20:30]
        assert_dict_arrays_equal(result, data)

    def test_multidim_set_some_ranges(self):
        # Create array.
        uri = self.path("dense_write_subarray_multidim_set_some_ranges")
        schema = tiledb.ArraySchema(
            tiledb.Domain(
                tiledb.Dim(name="d1", domain=(0, 99), tile=100),
                tiledb.Dim(name="d2", domain=(0, 99), tile=100),
            ),
            [tiledb.Attr(name="a1", dtype=np.float64)],
        )
        tiledb.Array.create(uri, schema)

        # Write data.
        data = np.random.rand(1000)
        with tiledb.open(uri, "w") as array:
            subarray = tiledb.Subarray(array)
            subarray.add_dim_range(1, (11, 20))
            array.write_subarray(subarray, {"a1": data})

        # Check results.
        with tiledb.open(uri, "r") as array:
            nonempty = array.nonempty_domain()
            assert nonempty[0] == (0, 99)
            assert nonempty[1] == (11, 20)
            result = array[:, 11:21]
        assert_dict_arrays_equal(result, {"a1": data.reshape(100, 10)})

    def test_with_negative_domain(self):
        # Create array.
        uri = self.path("dense_write_subarray_by_labels")
        schema = tiledb.ArraySchema(
            tiledb.Domain(
                tiledb.Dim(name="d1", domain=(-100, 100), tile=201, dtype=np.int32)
            ),
            [tiledb.Attr(name="a1", dtype=np.float64)],
        )
        tiledb.Array.create(uri, schema)

        # Define the data.
        data = OrderedDict([("a1", np.random.rand(5))])

        # Write full data and label data
        with tiledb.open(uri, "w") as array:
            subarray = tiledb.Subarray(array)
            subarray.add_dim_range(0, (-2, 2))
            array.write_subarray(subarray, data["a1"])

        # Check results
        with tiledb.open(uri, "r") as array:
            nonempty = array.nonempty_domain()
            assert nonempty[0] == (-2, 2)
            result = array.multi_index[-2:2]

        assert_dict_arrays_equal(result, data)

    def test_with_labels(self):
        # Create array.
        uri = self.path("dense_write_subarray_with_labels")
        dim1 = tiledb.Dim(name="d1", domain=(0, 10), tile=11)
        dim2 = tiledb.Dim(name="d2", domain=(0, 10), tile=11)
        schema = tiledb.ArraySchema(
            tiledb.Domain(dim1, dim2),
            [tiledb.Attr(name="a1", dtype=np.float64)],
            dim_labels={
                0: {"l1": dim1.create_label_schema("increasing", np.int32)},
                1: {"l2": dim1.create_label_schema("decreasing", np.int32)},
            },
        )
        tiledb.Array.create(uri, schema)

        data = OrderedDict(
            [
                ("a1", np.random.rand(121).reshape(11, 11)),
                ("l1", np.arange(-5, 6)),
                ("l2", np.arange(5, -6, -1)),
            ]
        )

        # Write full data and label data
        with tiledb.open(uri, "w") as array:
            subarray = tiledb.Subarray(array)
            subarray.add_dim_range(0, (0, 10))
            subarray.add_dim_range(1, (0, 10))
            array.write_subarray(subarray, data)

        # Check results
        with tiledb.open(uri, "r") as array:
            result = array.label_index(["l1", "l2"])[-5:5, -5:5]
        assert_dict_arrays_equal(result, data)

    def test_by_labels(self):
        # Create array.
        uri = self.path("dense_write_subarray_by_labels")
        dim1 = tiledb.Dim(name="d1", domain=(0, 10), tile=11)
        schema = tiledb.ArraySchema(
            tiledb.Domain(dim1),
            [tiledb.Attr(name="a1", dtype=np.float64)],
            dim_labels={0: {"l1": dim1.create_label_schema("increasing", np.int32)}},
        )
        tiledb.Array.create(uri, schema)

        # Define the data.
        data = OrderedDict(
            [("a1", np.random.rand(5)), ("l1", np.arange(-5, 6, dtype=np.int32))]
        )

        # Reload to get the label uris and write the labels.
        schema = tiledb.ArraySchema.load(uri)
        with tiledb.open(schema.dim_label("l1").uri, mode="w") as array:
            array[:] = data["l1"]

        # Write full data and label data
        with tiledb.open(uri, "w") as array:
            subarray = tiledb.Subarray(array)
            subarray.add_label_range("l1", (-2, 2))
            with pytest.raises(tiledb.TileDBError):
                array.write_subarray(subarray, data["a1"])

    def test_with_var_label(self):
        # Create array.
        uri = self.path("dense_write_subarray_by_var_label")
        dim1 = tiledb.Dim(name="d1", domain=(0, 10), tile=11)
        schema = tiledb.ArraySchema(
            tiledb.Domain(dim1),
            [tiledb.Attr(name="a1", dtype=np.float64)],
            dim_labels={
                0: {"l1": dim1.create_label_schema("increasing", "U")},
            },
        )
        tiledb.Array.create(uri, schema)

        # Write array.
        data = OrderedDict(
            [
                ("a1", np.random.rand(5)),
                (
                    "l1",
                    np.array(
                        ["alpha", "beta", "gamma", "kappa", "sigma"], dtype=object
                    ),
                ),
            ]
        )
        with tiledb.open(uri, "w") as array:
            subarray = tiledb.Subarray(array)
            subarray.add_dim_range(0, (3, 7))
            array.write_subarray(subarray, data)

        # Check results.
        with tiledb.open(uri, "r") as array:
            nonempty = array.nonempty_domain()
            assert nonempty[0] == (3, 7)
            with tiledb.open(array.schema.dim_label("l1").uri, "r") as label_array:
                nonempty_label = label_array.nonempty_domain()
                assert nonempty_label[0] == (3, 7)
            array.label_index(["l1"])["alpha":"sigma"]



import warnings
from typing import ByteString, Optional

import tiledb.libtiledb as lt

from .ctx import Ctx, default_ctx


class Filestore:
    """
    Functions to set and get data to and from a TileDB Filestore Array.

    A Filestore Array may be created using `ArraySchema.from_file` combined
    with `Array.create`.

    :param str uri: The URI to the TileDB Fileshare Array
    :param tiledb.Ctx ctx: A TileDB context
    """

    def __init__(self, uri: str, ctx: Optional[Ctx] = None) -> None:
        self._ctx = ctx or default_ctx()
        self._filestore_uri = uri

    def write(self, buffer: ByteString, mime_type: str = "AUTODETECT") -> None:
        """
        Import data from an object that supports the buffer protocol to a Filestore Array.

        :param buffer ByteString: Data of type bytes, bytearray, memoryview, etc.
        :param str mime_type: MIME types are "AUTODETECT" (default), "image/tiff", "application/pdf"

        """
        try:
            buffer = memoryview(buffer)
        except TypeError:
            raise TypeError(
                "Unexpected buffer type: buffer must support buffer protocol"
            )

        if not isinstance(mime_type, str):
            raise TypeError(
                f"Unexpected mime_type type '{type(mime_type)}': expected str"
            )

        try:
            lt.Filestore._buffer_import(
                self._ctx,
                self._filestore_uri,
                buffer,
                lt.Filestore._mime_type_from_str(mime_type),
            )
        except Exception as e:
            raise (e)

    def read(self, offset: int = 0, size: int = -1) -> bytes:
        """
        :param int offset: Byte position to begin reading. Defaults to beginning of filestore.
        :param int size: Total number of bytes to read. Defaults to -1 which reads the entire filestore.
        :rtype: bytes
        :return: Data from the Filestore Array

        """
        if not isinstance(offset, int):
            raise TypeError(f"Unexpected offset type '{type(offset)}': expected int")

        if not isinstance(size, int):
            raise TypeError(f"Unexpected size type '{type(size)}': expected int")

        if size == -1:
            size = len(self)
        size = min(size, len(self) - offset)

        return lt.Filestore._buffer_export(
            self._ctx,
            self._filestore_uri,
            offset,
            size,
        )

    @staticmethod
    def copy_from(
        filestore_array_uri: str,
        file_uri: str,
        mime_type: str = "AUTODETECT",
        ctx: Optional[Ctx] = None,
    ) -> None:
        """
        Copy data from a file to a Filestore Array.

        :param str filestore_array_uri: The URI to the TileDB Fileshare Array
        :param str file_uri: URI of file to export
        :param str mime_type: MIME types are "AUTODETECT" (default), "image/tiff", "application/pdf"
        :param tiledb.Ctx ctx: A TileDB context

        """
        if not isinstance(filestore_array_uri, str):
            raise TypeError(
                f"Unexpected filestore_array_uri type '{type(filestore_array_uri)}': expected str"
            )

        if not isinstance(file_uri, str):
            raise TypeError(
                f"Unexpected file_uri type '{type(file_uri)}': expected str"
            )

        if not isinstance(mime_type, str):
            raise TypeError(
                f"Unexpected mime_type type '{type(mime_type)}': expected str"
            )

        ctx = ctx or default_ctx()

        lt.Filestore._uri_import(
            ctx,
            filestore_array_uri,
            file_uri,
            lt.Filestore._mime_type_from_str(mime_type),
        )

    @staticmethod
    def copy_to(
        filestore_array_uri: str, file_uri: str, ctx: Optional[Ctx] = None
    ) -> None:
        """
        Copy data from a Filestore Array to a file.

        :param str filestore_array_uri: The URI to the TileDB Fileshare Array
        :param str file_uri: The URI to the TileDB Fileshare Array
        :param tiledb.Ctx ctx: A TileDB context

        """
        if not isinstance(filestore_array_uri, str):
            raise TypeError(
                f"Unexpected filestore_array_uri type '{type(filestore_array_uri)}': expected str"
            )

        if not isinstance(file_uri, str):
            raise TypeError(
                f"Unexpected file_uri type '{type(file_uri)}': expected str"
            )

        ctx = ctx or default_ctx()

        lt.Filestore._uri_export(ctx, filestore_array_uri, file_uri)

    def __len__(self) -> int:
        """
        :rtype: int
        :return: Bytes in the Filestore Array

        """
        return lt.Filestore._size(self._ctx, self._filestore_uri)



# object.py
#
# LICENSE
#
# The MIT License
#
# Copyright (c) 2020 TileDB, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# DESCRIPTION
#
# Please see the TileDB documentation for more information:
#   https://docs.tiledb.com/main/how-to/object-management
#
# This program creates a hierarchy as shown below. Specifically, it creates
# groups `dense_arrays` and `sparse_arrays` in a group `my_group`, and
# then some dense/sparse arrays and key-value store in those groups.
#
# my_group/
#   - dense_arrays/
#     - array_A
#     - array_B
#   - sparse_arrays/
#     - array_C
#     - array_D
#
# The program then shows how to list this hierarchy, as well as
# move/remove TileDB objects.

import os

import numpy as np

import tiledb


def create_array(array_name, sparse):
    if tiledb.object_type(array_name) == "array":
        return

    dom = tiledb.Domain(
        tiledb.Dim(name="rows", domain=(1, 4), tile=4, dtype=np.int32),
        tiledb.Dim(name="cols", domain=(1, 4), tile=4, dtype=np.int32),
    )
    schema = tiledb.ArraySchema(
        domain=dom, sparse=sparse, attrs=[tiledb.Attr(name="a", dtype=np.int32)]
    )
    if sparse:
        tiledb.SparseArray.create(array_name, schema)
    else:
        tiledb.DenseArray.create(array_name, schema)


def path(p):
    return os.path.join(os.getcwd(), p)


def create_hierarchy():
    # Create groups
    tiledb.group_create(path("my_group"))
    tiledb.group_create(path("my_group/dense_arrays"))
    tiledb.group_create(path("my_group/sparse_arrays"))

    # Create arrays
    create_array(path("my_group/dense_arrays/array_A"), False)
    create_array(path("my_group/dense_arrays/array_B"), False)
    create_array(path("my_group/sparse_arrays/array_C"), True)
    create_array(path("my_group/sparse_arrays/array_D"), True)


def list_obj(path):

    # List children
    print("\nListing hierarchy:")
    tiledb.ls(path, lambda obj_path, obj_type: print(obj_path, obj_type))

    # Walk in a path with a pre- and post-order traversal
    print("\nPreorder traversal:")
    tiledb.walk(
        path, lambda obj_path, obj_type: print(obj_path, obj_type)
    )  # Default order is preorder

    print("\nPostorder traversal:")
    tiledb.walk(
        path, lambda obj_path, obj_type: print(obj_path, obj_type), order="postorder"
    )


def move_remove_obj():
    tiledb.move(path("my_group"), path("my_group_2"))
    tiledb.remove(path("my_group_2/dense_arrays"))
    tiledb.remove(path("my_group_2/sparse_arrays/array_C"))


create_hierarchy()
list_obj("my_group")
move_remove_obj()  # Renames 'my_group' to 'my_group_2'
list_obj("my_group_2")

# clean up
tiledb.remove("my_group_2")



import io
import sys
from contextlib import contextmanager
from contextvars import ContextVar
from typing import Union

import tiledb
import tiledb.libtiledb as lt

_ctx_var = ContextVar("ctx")

already_warned = False
_needs_fork_wrapper = sys.platform != "win32" and sys.version_info < (3, 12)


class Config(lt.Config):
    """TileDB Config class

    The Config object stores configuration parameters for both TileDB Embedded
    and TileDB-Py.

    For TileDB Embedded parameters, see:

        https://docs.tiledb.com/main/how-to/configuration#configuration-parameters

    The following configuration options are supported by TileDB-Py:

        - `py.init_buffer_bytes`:

           Initial allocation size in bytes for attribute and dimensions buffers.
           If result size exceed the pre-allocated buffer(s), then the query will return
           incomplete and TileDB-Py will allocate larger buffers and resubmit.
           Specifying a sufficiently large buffer size will often improve performance.
           Default 10 MB (1024**2 * 10).

        - `py.use_arrow`:

           Use `pyarrow` from the Apache Arrow project to convert
           query results into Pandas dataframe format when requested.
           Default `True`.

        - `py.deduplicate`:

           Attempt to deduplicate Python objects during buffer
           conversion to Python. Deduplication may reduce memory usage for datasets
           with many identical strings, at the cost of some performance reduction
           due to hash calculation/lookup for each object.

    Unknown parameters will be ignored!

    :param dict params: Set parameter values from dict like object
    :param str path: Set parameter values from persisted Config parameter file
    """

    def __init__(self, params: dict = None, path: str = None):
        super().__init__()
        if path is not None:
            self.load(path)
        if params is not None:
            self.update(params)

    @staticmethod
    def load(uri: str):
        """Constructs a Config class instance from config parameters loaded from a local Config file

        :parameter str uri: a local URI config file path
        :rtype: tiledb.Config
        :return: A TileDB Config instance with persisted parameter values
        :raises TypeError: `uri` cannot be converted to a unicode string
        :raises: :py:exc:`tiledb.TileDBError`

        """
        return lt.Config(uri)

    def __setitem__(self, key: str, value: str):
        """Sets a config parameter value.

        :param str key: Name of parameter to set
        :param str value: Value of parameter to set
        :raises TypeError: `key` or `value` cannot be encoded into a UTF-8 string
        :raises: :py:exc:`tiledb.TileDBError`

        """
        self.set(str(key), str(value))

    def get(self, key: str, raise_keyerror: bool = True):
        try:
            return super().get(key)
        except Exception:
            if raise_keyerror:
                raise KeyError(key)
            else:
                return None

    def __getitem__(self, key: str):
        """Gets a config parameter value.

        :param str key: Name of parameter to get
        :return: Config parameter value string
        :rtype str:
        :raises TypeError: `key` cannot be encoded into a UTF-8 string
        :raises KeyError: Config parameter not found
        :raises: :py:exc:`tiledb.TileDBError`

        """
        return self.get(key, True)

    def __delitem__(self, key: str):
        """
        Removes a configured parameter (resetting it to its default).

        :param str key: Name of parameter to reset.
        :raises TypeError: `key` cannot be encoded into a UTF-8 string

        """
        self.unset(key)

    def __iter__(self):
        """Returns an iterator over the Config parameters (keys)"""
        return ConfigKeys(self)

    def __len__(self):
        """Returns the number of parameters (keys) held by the Config object"""
        return sum(1 for _ in self)

    def __eq__(self, config):
        if not isinstance(config, Config):
            return False
        keys = set(self.keys())
        okeys = set(config.keys())
        if keys != okeys:
            return False
        for k in keys:
            val, oval = self[k], config[k]
            if val != oval:
                return False
        return True

    unserialized_params_ = {
        "vfs.azure.storage_account_name",
        "vfs.azure.storage_account_key",
        "vfs.azure.storage_sas_token",
        "vfs.s3.proxy_username",
        "vfs.s3.proxy_password",
        "vfs.s3.aws_access_key_id",
        "vfs.s3.aws_secret_access_key",
        "vfs.s3.aws_session_token",
        "vfs.s3.aws_role_arn",
        "vfs.s3.aws_external_id",
        "vfs.s3.aws_load_frequency",
        "vfs.s3.aws_session_name",
        "vfs.gcs.service_account_key",
        "vfs.gcs.workload_identity_configuration",
        "vfs.gcs.impersonate_service_account",
        "rest.username",
        "rest.password",
        "rest.token",
    }

    def __repr__(self):
        colnames = ["Parameter", "Value"]
        params = list(self.keys())
        values = list(map(repr, self.values()))
        # for unserialized params, we don't want to print their values
        values = [
            "*" * 10 if p in self.unserialized_params_ and v != "''" else v
            for i, (p, v) in enumerate(zip(params, values))
        ]
        colsizes = [
            max(len(colnames[0]), *map(len, (p for p in params))),
            max(len(colnames[1]), *map(len, (v for v in values))),
        ]
        format_str = " | ".join("{{:<{}}}".format(i) for i in colsizes)
        output = []
        output.append(format_str.format(colnames[0], colnames[1]))
        output.append(format_str.format("-" * colsizes[0], "-" * colsizes[1]))
        output.extend(format_str.format(p, v) for p, v in zip(params, values))
        return "\n".join(output)

    def _repr_html_(self):
        output = io.StringIO()

        output.write("<table>")

        output.write("<tr>")
        output.write("<th>Parameter</th>")
        output.write("<th>Value</th>")
        output.write("</tr>")

        params = list(self.keys())
        values = list(map(repr, self.values()))
        # for unserialized params, we don't want to print their values
        values = [
            "*" * 10 if p in self.unserialized_params_ and v != "''" else v
            for i, (p, v) in enumerate(zip(params, values))
        ]

        for p, v in zip(params, values):
            output.write("<tr>")
            output.write(f"<td>{p}</td>")
            output.write(f"<td>{v}</td>")
            output.write("</tr>")

        output.write("</table>")

        return output.getvalue()

    def items(self, prefix: str = ""):
        """Returns an iterator object over Config parameters, values

        :param str prefix: return only parameters with a given prefix
        :rtype: ConfigItems
        :returns: iterator over Config parameter, value tuples

        """
        return ConfigItems(self, prefix=prefix)

    def keys(self, prefix: str = ""):
        """Returns an iterator object over Config parameters (keys)

        :param str prefix: return only parameters with a given prefix
        :rtype: ConfigKeys
        :returns: iterator over Config parameter string keys

        """
        return ConfigKeys(self, prefix=prefix)

    def values(self, prefix: str = ""):
        """Returns an iterator object over Config values

        :param str prefix: return only parameters with a given prefix
        :rtype: ConfigValues
        :returns: iterator over Config string values

        """
        return ConfigValues(self, prefix=prefix)

    def dict(self, prefix: str = ""):
        """Returns a dict representation of a Config object

        :param str prefix: return only parameters with a given prefix
        :rtype: dict
        :return: Config parameter / values as a Python dict

        """
        return dict(ConfigItems(self, prefix=prefix))

    def clear(self):
        """Unsets all Config parameters (returns them to their default values)"""
        for k in self.keys():
            del self[k]

    # def get(self, key, *args: Optional[str]):
    #     """Gets the value of a config parameter, or a default value.

    #     :param str key: Config parameter
    #     :param args: return `arg` if Config does not contain parameter `key`
    #     :return: Parameter value, `arg` or None.

    #     """
    #     nargs = len(args)
    #     if nargs > 1:
    #         raise TypeError("get expected at most 2 arguments, got {}".format(nargs))
    #     try:
    #         return self[key]
    #     except KeyError:
    #         return args[0] if nargs == 1 else None

    def update(self, odict: dict):
        """Update a config object with parameter, values from a dict like object

        :param odict: dict-like object containing parameter, values to update Config.

        """
        super().update(dict(odict))

    def from_file(self, path: str):
        """Update a Config object with from a persisted config file

        :param path: A local Config file path

        """
        config = Config.load(path)
        self.update(config)

    def save(self, uri: str):
        """Persist Config parameter values to a config file

        :parameter str uri: a local URI config file path
        :raises TypeError: `uri` cannot be converted to a unicode string
        :raises: :py:exc:`tiledb.TileDBError`

        """
        self.save_to_file(uri)

    def __reduce__(self):
        """
        Customize the pickling process by defining how to serialize
        and reconstruct the Config object.
        """
        state = self.dict()
        return (self.__class__, (), state)

    def __setstate__(self, state):
        """
        Customize how the Config object is restored from its serialized state.
        """
        self.__init__(state)


class ConfigKeys:
    """
    An iterator object over Config parameter strings (keys)
    """

    def __init__(self, config: Config, prefix: str = ""):
        self.config_items = ConfigItems(config, prefix=prefix)

    def __iter__(self):
        return self

    def __next__(self):
        (k, _) = self.config_items.__next__()
        return k


class ConfigValues:
    """
    An iterator object over Config parameter value strings
    """

    def __init__(self, config: Config, prefix: str = ""):
        self.config_items = ConfigItems(config, prefix=prefix)

    def __iter__(self):
        return self

    def __next__(self):
        (_, v) = self.config_items.__next__()
        return v


class ConfigItems:
    """
    An iterator object over Config parameter, values

    :param config: TileDB Config object
    :type config: tiledb.Config
    :param prefix: (default "") Filter paramter names with given prefix
    :type prefix: str

    """

    def __init__(self, config: Config, prefix: str = ""):
        self.config = config
        self.iter = config._iter(prefix)

    def __iter__(self):
        return self.iter

    def __next__(self):
        return self.iter.__next__()


class Ctx(lt.Context):
    """Class representing a TileDB context.

    A TileDB context wraps a TileDB storage manager.

    :param config: Initialize Ctx with given config parameters
    :type config: tiledb.Config or dict

    """

    def __init__(self, config: Config = None):
        _config = lt.Config()

        if config is not None:
            if isinstance(config, lt.Config):
                _config = config
            elif isinstance(config, Config):
                _config.update(config.dict())
            elif isinstance(config, dict):
                _config.update(config)
            else:
                raise TypeError(
                    "Ctx's config argument expects type `tiledb.Config` or `dict`"
                )

        super().__init__(_config)

        self._set_default_tags()

        # The core tiledb library uses threads and it's easy
        # to experience deadlocks when forking a process that is using
        # tiledb.  The project doesn't have a solution for this at the
        # moment other than to avoid using fork(), which is the same
        # recommendation that Python makes. Python 3.12 warns if you
        # fork() when multiple threads are detected and Python 3.14 will
        # make it so you never accidentally fork(): multiprocessing will
        # default to "spawn" on Linux.
        _ensure_os_fork_wrap()

    def __repr__(self):
        return "tiledb.Ctx() [see Ctx.config() for configuration]"

    def config(self):
        """Returns the Config instance associated with the Ctx."""
        new = Config.__new__(Config)
        # bypass calling Config.__init__, call lt.Config.__init__ instead
        lt.Config.__init__(new, super().config())
        return new

    def set_tag(self, key: str, value: str):
        """Sets a (string, string) "tag" on the Ctx (internal)."""
        super().set_tag(key, value)

    def _set_default_tags(self):
        """Sets all default tags on the Ctx"""
        self.set_tag("x-tiledb-api-language", "python")
        self.set_tag(
            "x-tiledb-api-language-version",
            f"{sys.version_info.major}."
            f"{sys.version_info.minor}."
            f"{sys.version_info.micro}",
        )
        self.set_tag("x-tiledb-api-sys-platform", sys.platform)

    def get_stats(self, print_out: bool = True, json: bool = False):
        """Retrieves the stats from a TileDB context.

        :param print_out: Print string to console (default True), or return as string
        :param json: Return stats JSON object (default: False)
        """
        stats = super().get_stats()

        if json:
            import json

            output = json.loads(stats)
        else:
            output = stats

        if print_out:
            print(output)
        else:
            return output


class CtxMixin:
    """
    Base mixin class for pure Python classes that extend PyBind11 TileDB classes.

    To use this class, a subclass must:
    - Inherit from it first (i.e. `class Foo(CtxMixin, Bar)`, not `class Foo(Bar, CtxMixin)`
    - Call super().__init__ by passing `ctx` (tiledb.Ctx or None) as first parameter and
      zero or more pure Python positional parameters
    """

    def __init__(self, ctx, *args, _pass_ctx_to_super=True):
        if not ctx:
            ctx = default_ctx()

        if _pass_ctx_to_super:
            super().__init__(ctx, *args)
        else:
            super().__init__(*args)

        # we set this here because if the super().__init__() constructor above fails,
        # we don't want to set self._ctx
        self._ctx = ctx

    @classmethod
    def from_capsule(cls, ctx, capsule):
        """Create an instance of this class from a PyCapsule instance"""
        # bypass calling self.__init__, call CtxMixin.__init__ instead
        self = cls.__new__(cls)
        CtxMixin.__init__(self, ctx, capsule)
        return self

    @classmethod
    def from_pybind11(cls, ctx, lt_obj):
        """Create an instance of this class from a PyBind11 instance"""
        # bypass calling self.__init__, call CtxMixin.__init__ instead
        self = cls.__new__(cls)
        CtxMixin.__init__(self, ctx, lt_obj, _pass_ctx_to_super=False)
        return self


def check_ipykernel_warn_once():
    """
    This function checks if we have imported ipykernel version < 6 in the
    current process, and provides a warning that default_ctx/scope_ctx will
    not work correctly due to a bug in IPython contextvar support."""
    global already_warned
    if not already_warned:
        try:
            import warnings

            if "ipykernel" in sys.modules and tuple(
                map(int, sys.modules["ipykernel"].__version__.split("."))
            ) < (6, 0):
                warnings.warn(
                    "tiledb.default_ctx and scope_ctx will not function correctly "
                    "due to bug in IPython contextvar support.  You must supply a "
                    "Ctx object to each function for custom configuration options. "
                    "Please consider upgrading to ipykernel >= 6!"
                    "Please see https://github.com/TileDB-Inc/TileDB-Py/issues/667 "
                    "for more information."
                )
        except:
            pass
        finally:
            already_warned = True


@contextmanager
def scope_ctx(ctx_or_config: Union["Ctx", "Config", dict] = None) -> "Ctx":
    """
    Context manager for setting the default `tiledb.Ctx` context variable when entering
    a block of code and restoring it to its previous value when exiting the block.

    :param ctx_or_config: :py:class:`tiledb.Ctx` or :py:class:`tiledb.Config` object
        or dictionary with config parameters.
    :return: Ctx
    """
    check_ipykernel_warn_once()

    if ctx_or_config is not None and not (
        isinstance(ctx_or_config, tiledb.Ctx)
        or isinstance(ctx_or_config, tiledb.Config)
        or isinstance(ctx_or_config, dict)
    ):
        raise ValueError(
            "scope_ctx takes in `tiledb.Ctx` object, `tiledb.Config` object, or "
            "dictionary with config parameters."
        )

    if not isinstance(ctx_or_config, tiledb.Ctx):
        ctx = tiledb.Ctx(ctx_or_config)
    else:
        ctx = ctx_or_config
    token = _ctx_var.set(ctx)
    try:
        yield _ctx_var.get()
    finally:
        _ctx_var.reset(token)


def default_ctx(config: Union["Config", dict] = None) -> "Ctx":
    """
    Returns, and optionally initializes, the default `tiledb.Ctx` context variable.

    This Ctx object is used by Python API functions when no `ctx` keyword argument
    is provided. Most API functions accept an optional `ctx` kwarg, but that is typically
    only necessary in advanced usage with multiple contexts per program.

    For initialization, this function must be called before any other tiledb functions.
    The initialization call accepts a  :py:class:`tiledb.Config` object to override the
    defaults for process-global parameters.

    :param config: :py:class:`tiledb.Config` object or dictionary with config parameters.
    :return: Ctx
    """
    check_ipykernel_warn_once()

    if config is not None and not (
        isinstance(config, tiledb.Config) or isinstance(config, dict)
    ):
        raise ValueError(
            "default_ctx takes in `tiledb.Config` object or "
            "dictionary with config parameters."
        )

    try:
        ctx = _ctx_var.get()
        if config is not None:
            raise tiledb.TileDBError("Global context already initialized!")

        # The core tiledb library uses threads and it's easy
        # to experience deadlocks when forking a process that is using
        # tiledb.  The project doesn't have a solution for this at the
        # moment other than to avoid using fork(), which is the same
        # recommendation that Python makes. Python 3.12 warns if you
        # fork() when multiple threads are detected and Python 3.14 will
        # make it so you never accidentally fork(): multiprocessing will
        # default to "spawn" on Linux.
        _ensure_os_fork_wrap()
    except LookupError:
        ctx = tiledb.Ctx(config)
        _ctx_var.set(ctx)
    return ctx


def _ensure_os_fork_wrap():
    global _needs_fork_wrapper
    if _needs_fork_wrapper:
        import os
        import warnings
        from functools import wraps

        def warning_wrapper(func):
            @wraps(func)
            def wrapper():
                warnings.warn(
                    "TileDB is a multithreading library and deadlocks "
                    "are likely if fork() is called after a TileDB "
                    "context has been created (such as for array "
                    "access). To safely use TileDB with "
                    "multiprocessing or concurrent.futures, choose "
                    "'spawn' as the start method for child processes. "
                    "For example: "
                    "multiprocessing.set_start_method('spawn').",
                    UserWarning,
                )
                return func()

            return wrapper

        os.fork = warning_wrapper(os.fork)
        _needs_fork_wrapper = False



import importlib.util
import json
import time
import weakref
from abc import ABC, abstractmethod
from collections import OrderedDict
from contextlib import contextmanager
from contextvars import ContextVar
from dataclasses import dataclass
from numbers import Real
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
)

import numpy as np

from .aggregation import Aggregation as AggregationProxy
from .array import Array
from .array_schema import ArraySchema
from .dataframe_ import check_dataframe_deps
from .libtiledb import TileDBError
from .main import PyAgg, PyQuery, increment_stat, use_stats
from .metadata import Metadata
from .query import Query
from .query import Query as QueryProxy
from .query_condition import QueryCondition
from .subarray import Subarray

if TYPE_CHECKING:
    # We don't want to import these eagerly since importing Pandas in particular
    # can add around half a second of import time even if we never use it.
    import pandas
    import pyarrow


current_timer: ContextVar[str] = ContextVar("timer_scope")


# sentinel type to denote selecting an empty range
EmptyRange = TypeVar("EmptyRange")


def is_empty_range(idx: Union[EmptyRange, List, Tuple]) -> bool:
    if idx is not EmptyRange:
        if hasattr(idx, "__len__") and len(idx) == 0 and idx != "":
            return True
        return False
    return True


# TODO: expand with more accepted scalar types
Scalar = Real
Range = Tuple[Scalar, Scalar]


@dataclass
class EstimatedResultSize:
    offsets_bytes: int
    data_bytes: int


@contextmanager
def timing(key: str) -> Iterator[None]:
    if not use_stats():
        yield
    else:
        scoped_name = f"{current_timer.get('py')}.{key}"
        parent_token = current_timer.set(scoped_name)
        start = time.time()
        try:
            yield
        finally:
            increment_stat(current_timer.get(), time.time() - start)
            current_timer.reset(parent_token)


def mr_dense_result_shape(
    ranges: Sequence[Sequence[Range]], base_shape: Optional[Tuple[int, ...]] = None
) -> Tuple[int, ...]:
    if base_shape is not None:
        assert len(ranges) == len(base_shape), "internal error: mismatched shapes"

    new_shape = []
    for i, subranges in enumerate(ranges):
        if isinstance(subranges, np.ndarray):
            total_length = len(subranges)
            new_shape.append(np.uint64(total_length))
        elif subranges not in (None, ()):
            total_length = sum(abs(stop - start) + 1 for start, stop in subranges)
            new_shape.append(np.uint64(total_length))
        elif base_shape is not None:
            # empty range covers dimension
            new_shape.append(base_shape[i])
        else:
            raise ValueError("Missing required base_shape for whole-dimension slices")

    return tuple(new_shape)


def to_scalar(obj: Any) -> Scalar:
    if np.isscalar(obj):
        return cast(Scalar, obj)
    if isinstance(obj, np.ndarray) and obj.ndim == 0:
        return cast(Scalar, obj[()])
    raise ValueError(f"Cannot convert {type(obj)} to scalar")


def iter_ranges(
    sel: Union[Scalar, slice, Range, List[Scalar]],
    sparse: bool,
    nonempty_domain: Optional[Range] = None,
) -> Iterator[Range]:
    if isinstance(sel, slice):
        if sel.step is not None:
            raise ValueError("Stepped slice ranges are not supported")

        rstart = sel.start
        if rstart is None and nonempty_domain:
            rstart = nonempty_domain[0]

        rend = sel.stop
        if rend is None and nonempty_domain:
            rend = nonempty_domain[1]

        if sparse and sel.start is None and sel.stop is None:
            # don't set nonempty_domain for full-domain slices w/ sparse
            # because TileDB query is faster without the constraint
            pass
        elif rstart is None or rend is None:
            pass
        else:
            yield to_scalar(rstart), to_scalar(rend)

    elif isinstance(sel, tuple):
        assert len(sel) == 2
        yield to_scalar(sel[0]), to_scalar(sel[1])

    elif isinstance(sel, list):
        for scalar in map(to_scalar, sel):
            yield scalar, scalar

    else:
        scalar = to_scalar(sel)
        yield scalar, scalar


def iter_label_range(sel: Union[Scalar, slice, Range, List[Scalar]]):
    if isinstance(sel, slice):
        if sel.start is None or sel.start is None:
            raise NotImplementedError(
                "partial and full indexing is not yet supported on dimension labels"
            )

        yield to_scalar(sel.start), to_scalar(sel.stop)

    elif isinstance(sel, tuple):
        assert len(sel) == 2
        yield to_scalar(sel[0]), to_scalar(sel[1])

    elif isinstance(sel, list):
        for scalar in map(to_scalar, sel):
            yield scalar, scalar

    else:
        scalar = to_scalar(sel)
        yield scalar, scalar


def dim_ranges_from_selection(selection, nonempty_domain, is_sparse):
    # don't try to index nonempty_domain if None
    if isinstance(selection, np.ndarray):
        return selection
    selection = selection if isinstance(selection, list) else [selection]
    return tuple(
        rng for sel in selection for rng in iter_ranges(sel, is_sparse, nonempty_domain)
    )


def label_ranges_from_selection(selection):
    if isinstance(selection, np.ndarray):
        return tuple(tuple(x, x) for x in selection)
    selection = selection if isinstance(selection, list) else [selection]
    return tuple(rng for sel in selection for rng in iter_label_range(sel))


def getitem_ranges(array: Array, idx: Any) -> Sequence[Sequence[Range]]:
    ranges: List[Sequence[Range]] = [()] * array.schema.domain.ndim
    ned = array.nonempty_domain()
    if ned is None:
        ned = [None] * array.schema.domain.ndim
    is_sparse = array.schema.sparse
    for i, dim_sel in enumerate([idx] if not isinstance(idx, tuple) else idx):
        ranges[i] = dim_ranges_from_selection(dim_sel, ned[i], is_sparse)
    return tuple(ranges)


def getitem_ranges_with_labels(
    array: Array, labels: Dict[int, str], idx: Any
) -> Tuple[Sequence[Sequence[Range]], Dict[str, Sequence[Range]]]:
    dim_ranges: List[Sequence[Range]] = [()] * array.schema.domain.ndim
    label_ranges: Dict[str, Sequence[Range]] = {}
    ned = array.nonempty_domain()
    if ned is None:
        ned = [None] * array.schema.domain.ndim
    is_sparse = array.schema.sparse
    for dim_idx, dim_sel in enumerate([idx] if not isinstance(idx, tuple) else idx):
        if dim_idx in labels.keys():
            label_ranges[labels[dim_idx]] = label_ranges_from_selection(dim_sel)
        else:
            dim_ranges[dim_idx] = dim_ranges_from_selection(
                dim_sel, ned[dim_idx], is_sparse
            )
    return dim_ranges, label_ranges


class _BaseIndexer(ABC):
    """
    Implements multi-range indexing.
    """

    def __init__(
        self,
        array: Array,
        query: Optional[QueryProxy] = None,
        use_arrow: bool = False,
        preload_metadata: bool = False,
    ):
        if not isinstance(array, Array):
            raise TypeError("_BaseIndexer expected tiledb.Array")
        self.array_ref = weakref.ref(array)
        self.query = query
        self.use_arrow = use_arrow
        self.preload_metadata = preload_metadata
        self.subarray = None
        self.pyquery = None

    @property
    def array(self) -> Array:
        array = self.array_ref()
        if array is None:
            raise RuntimeError(
                "Internal error: invariant violation (indexing call w/ dead array_ref)"
            )
        return array

    @property
    def return_incomplete(self) -> bool:
        return bool(self.query and self.query.return_incomplete)

    def __getitem__(self, idx):
        with timing("getitem_time"):
            if is_empty_range(idx):
                self.pyquery = None
                self.subarray = None
            else:
                self._set_pyquery()
                self.subarray = Subarray(self.array)
                self._set_ranges(idx)
            return self if self.return_incomplete else self._run_query()

    def estimated_result_sizes(self):
        """
        Get the estimated result buffer sizes for a TileDB Query

        Sizes are returned in bytes as an EstimatedResultSize dataclass
        with two fields: `offset_bytes` and `data_bytes`, with buffer
        name as the OrderedDict key.
        See the corresponding TileDB Embedded API documentation for
        additional details:

        https://tiledb-inc-tiledb.readthedocs-hosted.com/en/stable/c++-api.html#query

        :return: OrderedDict of key: str -> EstimatedResultSize
        """
        if not hasattr(self, "pyquery"):
            raise TileDBError("Query not initialized")

        if self.pyquery is None:
            return {
                name: EstimatedResultSize(0, 0) for name in self._empty_results.keys()
            }
        else:
            return {
                name: EstimatedResultSize(*values)
                for name, values in self.pyquery.estimated_result_sizes().items()
            }

    def __iter__(self):
        if not hasattr(self, "pyquery"):
            raise TileDBError("Query not initialized")
        if not self.return_incomplete:
            raise TileDBError(
                "Cannot iterate unless query is initialized with return_incomplete=True"
            )
        while True:
            yield self._run_query()
            if self.pyquery is None or not self.pyquery.is_incomplete:
                break

    @property
    def _empty_results(self):
        return _get_empty_results(self.array.schema, self.query)

    def _set_pyquery(self):
        self.pyquery = _get_pyquery(
            self.array,
            self.query,
            self.use_arrow,
            self.return_incomplete,
            self.preload_metadata,
        )

    def _set_ranges(self, idx):
        ranges = getitem_ranges(self.array, idx)
        self._set_shape(ranges)
        with timing("add_ranges"):
            self.subarray.add_ranges(ranges)
        self.pyquery.set_subarray(self.subarray)

    def _set_shape(self, ranges):
        pass

    @abstractmethod
    def _run_query(self):
        """Run the query for the latest __getitem__ call and return the result"""


class MultiRangeIndexer(_BaseIndexer):
    """
    Implements multi-range indexing.
    """

    def __init__(self, array: Array, query: Optional[QueryProxy] = None):
        if query and query.return_arrow:
            raise TileDBError("`return_arrow=True` requires .df indexer`")
        super().__init__(array, query)
        self.result_shape = None

    def _set_shape(self, ranges):
        schema = self.array.schema
        if not schema.sparse and len(schema.shape) > 1:
            self.result_shape = mr_dense_result_shape(ranges, schema.shape)
        else:
            self.result_shape = None

    def _run_query(self) -> Dict[str, np.ndarray]:
        if self.pyquery is None:
            return self._empty_results

        self.pyquery.submit()
        result_dict = _get_pyquery_results(self.pyquery, self.array)
        if self.result_shape is not None:
            for name, arr in result_dict.items():
                # TODO check/test layout
                if not self.array.schema.has_dim_label(name):
                    arr.shape = self.result_shape
        return result_dict


class MultiRangeAggregation(_BaseIndexer):
    def __init__(self, array: Array, query: Optional[AggregationProxy] = None):
        super().__init__(array, query)
        self.result_shape = None

    def _set_shape(self, ranges):
        schema = self.array.schema
        if not schema.sparse and len(schema.shape) > 1:
            self.result_shape = mr_dense_result_shape(ranges, schema.shape)
        else:
            self.result_shape = None

    def __getitem__(self, idx):
        with timing("getitem_time"):
            if idx is EmptyRange:
                self.pyquery = None
                self.subarray = None
            else:
                self.pyquery = _get_pyagg(self.array, self.query)
                self.subarray = Subarray(self.array)
                self._set_ranges(idx)
            return self._run_query()

    def _run_query(self) -> Dict[str, np.ndarray]:
        if self.pyquery is None:
            return self._empty_results

        result = self.pyquery.get_aggregate()

        # If there was only one attribute, just show the aggregate results
        if len(result) == 1:
            result = result[list(result.keys())[0]]

            # If there was only one aggregate, just show the value
            if len(result) == 1:
                result = result[list(result.keys())[0]]

        return result


class DataFrameIndexer(_BaseIndexer):
    """
    Implements `.df[]` indexing to directly return a dataframe
    [] operator uses multi_index semantics.
    """

    def __init__(
        self,
        array: Array,
        query: Optional[QueryProxy] = None,
        use_arrow: Optional[bool] = None,
    ):
        check_dataframe_deps()
        # we need to use a Query in order to get coords for a dense array
        if not query:
            query = QueryProxy(array, has_coords=True)
        use_arrow = (
            bool(importlib.util.find_spec("pyarrow"))
            if use_arrow is None
            else use_arrow
        )

        # TODO: currently there is lack of support for Arrow list types. This prevents
        # multi-value attributes, asides from strings, from being queried properly.
        # Until list attributes are supported in core, error with a clear message.
        if use_arrow and any(
            (attr.isvar or len(attr.dtype) > 1)
            and attr.dtype not in (np.str_, np.bytes_)
            for attr in map(array.attr, query.attrs or ())
        ):
            raise TileDBError(
                "Multi-value attributes are not currently supported when use_arrow=True. "
                "This includes all variable-length attributes and fixed-length "
                "attributes with more than one value. Use `query(use_arrow=False)`."
            )
        super().__init__(array, query, use_arrow, preload_metadata=True)

    def _run_query(self) -> Union["pandas.DataFrame", "pyarrow.Table"]:
        import pandas
        import pyarrow

        if self.pyquery is not None:
            self.pyquery.submit()

        if self.pyquery is None:
            df = pandas.DataFrame(self._empty_results)
        elif self.use_arrow:
            with timing("buffer_conversion_time"):
                table = self.pyquery._buffers_to_pa_table()

            columns = []
            pa_schema = table.schema
            for pa_attr in pa_schema:
                if not self.array.schema.has_attr(pa_attr.name):
                    continue

                tdb_attr = self.array.attr(pa_attr.name)

                if tdb_attr.enum_label is not None:
                    enmr = self.array.enum(tdb_attr.enum_label)
                    col = pyarrow.DictionaryArray.from_arrays(
                        indices=table[pa_attr.name].combine_chunks(),
                        dictionary=enmr.values(),
                    )
                    idx = pa_schema.get_field_index(pa_attr.name)
                    table = table.set_column(idx, pa_attr.name, col)
                    pa_schema = table.schema
                    continue

                if np.issubdtype(tdb_attr.dtype, bool):
                    # this is a workaround to cast TILEDB_BOOL types from uint8
                    # representation in Arrow to Boolean
                    dtype = "uint8"
                elif tdb_attr.isnullable and np.issubdtype(tdb_attr.dtype, np.integer):
                    # this is a workaround for PyArrow's to_pandas function
                    # converting all integers with NULLs to float64:
                    # https://arrow.apache.org/docs/python/pandas.html#arrow-pandas-conversion
                    extended_dtype_mapping = {
                        pyarrow.int8(): pandas.Int8Dtype(),
                        pyarrow.int16(): pandas.Int16Dtype(),
                        pyarrow.int32(): pandas.Int32Dtype(),
                        pyarrow.int64(): pandas.Int64Dtype(),
                        pyarrow.uint8(): pandas.UInt8Dtype(),
                        pyarrow.uint16(): pandas.UInt16Dtype(),
                        pyarrow.uint32(): pandas.UInt32Dtype(),
                        pyarrow.uint64(): pandas.UInt64Dtype(),
                    }
                    dtype = extended_dtype_mapping[pa_attr.type]
                else:
                    continue

                columns.append(
                    {
                        "field_name": tdb_attr.name,
                        "name": tdb_attr.name,
                        "numpy_type": f"{dtype}",
                        "pandas_type": f"{dtype}",
                    }
                )

            metadata = {
                b"pandas": json.dumps(
                    {
                        "columns": columns,
                        "index_columns": [
                            {
                                "kind": "range",
                                "name": None,
                                "start": 0,
                                "step": 1,
                                "stop": len(table),
                            }
                        ],
                    }
                ).encode()
            }

            table = table.cast(pyarrow.schema(pa_schema).with_metadata(metadata))

            if self.query.return_arrow:
                return table

            df = table.to_pandas()
        else:
            df = pandas.DataFrame(_get_pyquery_results(self.pyquery, self.array))

        with timing("pandas_index_update_time"):
            return _update_df_from_meta(df, self.array.meta, self.query.index_col)


class LabelIndexer(MultiRangeIndexer):
    """
    Implements multi-range indexing by label.
    """

    def __init__(
        self, array: Array, labels: Sequence[str], query: Optional[QueryProxy] = None
    ):
        if array.schema.sparse:
            raise NotImplementedError(
                "querying sparse arrays by label is not yet implemented"
            )
        super().__init__(array, query)
        self.label_query: Optional[Query] = None
        self._labels: Dict[int, str] = {}
        for label_name in labels:
            dim_label = array.schema.dim_label(label_name)
            dim_idx = dim_label.dim_index
            if dim_idx in self._labels:
                raise TileDBError(
                    f"cannot set labels `{self._labels[dim_idx]}` and "
                    f"`{label_name}` defined on the same dimension"
                )
            self._labels[dim_idx] = label_name

    def _set_ranges(self, idx):
        dim_ranges, label_ranges = getitem_ranges_with_labels(
            self.array, self._labels, idx
        )
        if label_ranges is None:
            with timing("add_ranges"):
                self.subarray.add_ranges(tuple(dim_ranges))
            # No label query.
            self.label_query = None
            # All ranges are finalized: set shape and subarray now.
            self._set_shape(dim_ranges)
            self.pyquery.set_subarray(self.subarray)
        else:
            label_subarray = Subarray(self.array)
            with timing("add_ranges"):
                self.subarray.add_ranges(dim_ranges=dim_ranges)
                label_subarray.add_ranges(label_ranges=label_ranges)
            self.label_query = Query(self.array)
            self.label_query.set_subarray(label_subarray)

    def _run_query(self) -> Dict[str, np.ndarray]:
        # If querying by label and the label query is not yet complete, run the label
        # query and update the pyquery with the actual dimensions.
        if self.label_query is not None and not self.label_query.is_complete():
            self.label_query._submit()

            if not self.label_query.is_complete():
                raise TileDBError("failed to get dimension ranges from labels")
            label_subarray = self.label_query.subarray()
            # Check that the label query returned results for all dimensions.
            if any(
                label_subarray.num_dim_ranges(dim_idx) == 0 for dim_idx in self._labels
            ):
                self.pyquery = None
            else:
                # Get the ranges from the label query and set to the
                self.subarray.copy_ranges(
                    self.label_query.subarray(), self._labels.keys()
                )
                self.pyquery.set_subarray(self.subarray)
            self.result_shape = self.subarray.shape()
            for dim_idx, label_name in self._labels.items():
                if self.result_shape is None:
                    raise TileDBError("failed to compute subarray shape")
                self.pyquery.add_label_buffer(label_name, self.result_shape[dim_idx])
        return super()._run_query()


def _get_pyquery(
    array: Array,
    query: Optional[QueryProxy],
    use_arrow: bool,
    return_incomplete: bool,
    preload_metadata: bool,
) -> PyQuery:
    schema = array.schema
    if query:
        order = query.order
    else:
        # set default order:  TILEDB_UNORDERED for sparse,  TILEDB_ROW_MAJOR for dense
        order = "U" if schema.sparse else "C"

    try:
        layout = "CFGU".index(order)
    except ValueError:
        raise ValueError(
            "order must be 'C' (TILEDB_ROW_MAJOR), 'F' (TILEDB_COL_MAJOR),  "
            "'U' (TILEDB_UNORDERED), or 'G' (TILEDB_GLOBAL_ORDER)"
        )

    pyquery = PyQuery(
        array.ctx,
        array,
        tuple(
            [array.view_attr]
            if array.view_attr is not None
            else _iter_attr_names(schema, query)
        ),
        tuple(_iter_dim_names(schema, query)),
        layout,
        use_arrow,
    )

    pyquery._return_incomplete = return_incomplete
    pyquery._preload_metadata = preload_metadata
    if query and query.cond is not None:
        if isinstance(query.cond, str):
            pyquery.set_cond(QueryCondition(query.cond))
        else:
            raise TypeError("`cond` expects type str.")

    return pyquery


def _get_pyagg(array: Array, agg: AggregationProxy) -> PyAgg:
    order = agg.query.order

    try:
        layout = "CFGU".index(order)
    except ValueError:
        raise ValueError(
            "order must be 'C' (TILEDB_ROW_MAJOR), 'F' (TILEDB_COL_MAJOR),  "
            "'U' (TILEDB_UNORDERED), or 'G' (TILEDB_GLOBAL_ORDER)"
        )

    pyagg = PyAgg(array.ctx, array, layout, agg.attr_to_aggs)
    if agg.query.cond is not None:
        pyagg.set_cond(QueryCondition(agg.query.cond))
    return pyagg


def _iter_attr_names(
    schema: ArraySchema, query: Optional[QueryProxy] = None
) -> Iterator[str]:
    if query is not None and query.attrs is not None:
        return iter(query.attrs)
    return (schema.attr(i)._internal_name for i in range(schema.nattr))


def _iter_dim_names(
    schema: ArraySchema, query: Optional[QueryProxy] = None
) -> Iterator[str]:
    if query is not None:
        if query.dims is not None:
            return iter(query.dims or ())
        if query.has_coords is False:
            return iter(())
    if not schema.sparse:
        return iter(())
    dom = schema.domain
    return (dom.dim(i).name for i in range(dom.ndim))


def _get_pyquery_results(pyquery: PyQuery, array: Array) -> Dict[str, np.ndarray]:
    schema = array.schema
    result_dict = OrderedDict()
    for name, item in pyquery.results().items():
        if len(item[1]) > 0:
            arr = pyquery.unpack_buffer(name, item[0], item[1])
        else:
            arr = item[0]
            arr.dtype = (
                schema.attr_or_dim_dtype(name)
                if not schema.has_dim_label(name)
                else schema.dim_label(name).dtype
            )

        if schema.has_attr(name):
            enum_label = schema.attr(name).enum_label
            if enum_label is not None:
                values = array.enum(enum_label).values()
                arr = np.array([values[idx] for idx in arr])

        result_dict[name if name != "__attr" else ""] = arr
    return result_dict


def _get_empty_results(
    schema: ArraySchema, query: Optional[QueryProxy] = None
) -> Dict[str, np.ndarray]:
    names = []
    query_dims = frozenset(_iter_dim_names(schema, query))
    query_attrs = frozenset(_iter_attr_names(schema, query))

    # return dims first, if any
    dom = schema.domain
    for i in range(dom.ndim):
        dim = dom.dim(i).name
        # we need to also check if this is an attr for backward-compatibility
        if dim in query_dims or dim in query_attrs:
            names.append(dim)

    for i in range(schema.nattr):
        attr = schema.attr(i)._internal_name
        if attr in query_attrs:
            names.append(attr)

    result_dict = OrderedDict()
    for name in names:
        arr = np.array([], schema.attr_or_dim_dtype(name))
        result_dict[name if name != "__attr" else ""] = arr
    return result_dict


def _update_df_from_meta(
    df: "pandas.DataFrame",
    array_meta: Metadata,
    index_col: Union[List[str], bool, None] = True,
) -> "pandas.DataFrame":
    col_dtypes = {}
    if "__pandas_attribute_repr" in array_meta:
        attr_dtypes = json.loads(array_meta["__pandas_attribute_repr"])
        for name, dtype in attr_dtypes.items():
            if name in df:
                col_dtypes[name] = dtype

    index_cols = []
    if "__pandas_index_dims" in array_meta:
        index_dtypes = json.loads(array_meta["__pandas_index_dims"])
        index_cols.extend(col for col in index_dtypes.keys() if col in df)
        for name, dtype in index_dtypes.items():
            if name in df:
                col_dtypes[name] = dtype

    if col_dtypes:
        df = df.astype(col_dtypes, copy=False)

    if index_col:
        if index_col is not True:
            # if we have a query with index_col set, then override any
            # index information saved with the array.
            df.set_index(index_col, inplace=True)
        elif index_cols:
            # set index the index names that exist as columns
            df.set_index(index_cols, inplace=True)

            # rename __tiledb_rows to None
            if "__tiledb_rows" in index_cols:
                index_cols[index_cols.index("__tiledb_rows")] = None
                if len(index_cols) == 1:
                    df.index.rename(index_cols[0], inplace=True)
                else:
                    df.index.rename(index_cols, inplace=True)

    return df
